"""Definitions for review request fields."""

from __future__ import unicode_literals

import logging

from django.template.loader import render_to_string
from django.utils import six
from django.utils.functional import cached_property
from django.utils.html import escape, format_html_join, strip_tags
from django.utils.safestring import mark_safe
from django.utils.six.moves.html_parser import HTMLParser
from django.utils.translation import ugettext_lazy as _
from djblets.markdown import iter_markdown_lines
from djblets.registries.errors import ItemLookupError
from djblets.registries.registry import ALREADY_REGISTERED, NOT_REGISTERED

from reviewboard.diffviewer.diffutils import get_line_changed_regions
from reviewboard.diffviewer.myersdiff import MyersDiffer
from reviewboard.diffviewer.templatetags.difftags import highlightregion
from reviewboard.registries.registry import Registry, OrderedRegistry
from reviewboard.reviews.markdown_utils import (is_rich_text_default_for_user,
                                                normalize_text_for_edit,
                                                render_markdown)


class FieldSetRegistry(OrderedRegistry):
    """A registry for field sets.

    This keeps the fieldsets in the registered order, so iterating through them
    will do so in the same order.
    """

    lookup_attrs = ('fieldset_id',)

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered review request fieldset.'
        ),
        NOT_REGISTERED: _(
            '%(attr_value)s is not a registered review request fieldset.'
        ),
    }

    def __init__(self):
        """Initialize the registry."""
        self._key_order = []
        super(FieldSetRegistry, self).__init__()

    def register(self, fieldset):
        """Register the fieldset.

        This will also register all field classes registered on the fieldset on
        the field registry.

        Args:
            fieldset (type):
                The fieldset to register, as a
                :py:class:`BaseReviewRequestFieldSet` subclass.
        """
        super(FieldSetRegistry, self).register(fieldset)

        # Set the field_classes to an empty list by default if it doesn't
        # explicitly provide its own, so that entries don't go into
        # BaseReviewRequestFieldSet's global list.
        if fieldset.field_classes is None:
            fieldset.field_classes = []

        for field_cls in fieldset.field_classes:
            field_registry.register(field_cls)

    def unregister(self, fieldset):
        """Unregister the fieldset.

        This will unregister all field classes on the fieldset from the field
        registry.

        Args:
            fieldset (type):
                The field to remove, as a
                :py:class:`BaseReviewRequestFieldSet` subclass.
        """
        super(FieldSetRegistry, self).unregister(fieldset)

        for field_cls in fieldset.field_classes:
            fieldset.remove_field(field_cls)

    def get_defaults(self):
        """Return the list of built-in fieldsets.

        Returns:
            list:
            A list of the built-in
            :py:class:`~reviewboard.reviews.fields.BaseReviewRequestFieldSet`
            subclasses.
        """
        from reviewboard.reviews.builtin_fields import builtin_fieldsets

        return builtin_fieldsets


class FieldRegistry(Registry):
    """A registry for review request fields."""

    lookup_attrs = ['field_id']
    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered review request field. Field '
            'IDs must be unique across all fieldsets.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s is not a registered review request fieldset.'
        ),
    }

    def populate(self):
        """Populate the registry."""
        # Fields are only ever registered via the FieldSetRegistry, so we
        # ensure that it has been populated as well.
        fieldset_registry.populate()


fieldset_registry = FieldSetRegistry()
field_registry = FieldRegistry()


class BaseReviewRequestFieldSet(object):
    """Base class for sets of review request fields.

    A fieldset stores a list of fields that are rendered on the review
    request page. They may contain default fields, and new fields can be
    added or removed.

    Review Board provides three main fieldsets: "main", "info", and
    "reviewers". Others can be added by subclassing and registering
    through ``register_review_request_fieldset``.
    """

    fieldset_id = None
    label = None
    show_required = False
    field_classes = None
    tag_name = None

    def __init__(self, review_request_details):
        """Initialize the field set.

        Args:
            review_request_details (reviewboard.reviews.models.base_review_request_details.BaseReviewRequestDetails):
                The review request or draft.
        """
        self.review_request_details = review_request_details

    @classmethod
    def is_empty(cls):
        """Return whether the fieldset is empty.

        A fieldset is empty if there are no field classes registered.
        An empty fieldset will not be displayed on the page.
        """
        return not cls.field_classes

    @classmethod
    def add_field(cls, field_cls):
        """Add a field class to this fieldset.

        The field will be rendered inside this fieldset on the page.

        A given field class can only be in one fieldset. Its ``field_id``
        must be unique.

        Args:
            field_cls (class):
                The field class to add.
        """
        field_registry.register(field_cls)
        cls.field_classes.append(field_cls)

    @classmethod
    def remove_field(cls, field_cls):
        """Remove a field class from this fieldset.

        The field class must have been previously added to this fieldset.

        Args:
            field_cls (class):
                The field class to remove.
        """
        cls.field_classes.remove(field_cls)

        try:
            field_registry.unregister(field_cls)
        except ItemLookupError as e:
            logging.error('Failed to unregister unknown review request '
                          'field "%s"',
                          field_cls.field_id)
            raise e

    def __str__(self):
        """Represent the field set as a byte string.

        Returns:
            bytes:
            The field set's ID as a byte string.
        """
        if isinstance(self.fieldset_id, six.binary_type):
            return self.fieldset_id

        return self.fieldset_id.encode('utf-8')

    def __unicode__(self):
        """Represent the field set as a unicode string.

        Returns:
            unicode:
            The field set's ID as a unicode string.
        """
        if isinstance(self.fieldset_id, six.binary_type):
            return self.fieldset_id.decode('utf-8')

        return self.fieldset_id


class BaseReviewRequestField(object):
    """Base class for a field on a review request.

    A field is responsible for displaying data from a review request,
    handling any editing requirements if necessary, recording changes
    in the ChangeDescription, and rendering those changes.

    Each field must have its own unique ``field_id``. This ID will be used
    when looking up or storing the field's value.

    It is recommended that fields provided by extensions prefix their
    field ID with some sort of identifier for the extension or the vendor.

    Creating a new field requires subclassing BaseReviewRequestField and
    overriding any fields or functions necessary. Its class must then be
    added to a fieldset.

    A field will be instantiated with either a ReviewRequest or a
    ReviewRequestDraft, depending on what is being edited. This is stored
    in ``review_request_details``. Functions should optimistically fetch
    values from that, if possible. They can call ``get_review_request()``
    on ``review_request_details`` to fetch the actual ReviewRequest.

    If the function takes a ``review_request_details`` parameter, it must
    use that instead.
    """

    field_id = None
    label = None
    is_editable = False
    is_required = False
    default_css_classes = set()
    change_entry_renders_inline = True
    model = None

    #: The HTML tag to be used when rendering the field.
    tag_name = 'span'

    #: Whether the field should be rendered.
    should_render = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = None

    can_record_change_entry = property(lambda self: self.is_editable)

    def __init__(self, review_request_details, request=None):
        """Initialize the field.

        Args:
            review_request_details (reviewboard.reviews.models.base_review_request_details.BaseReviewRequestDetails):
                The review request or draft.

            request (django.http.HttpRequest, optional):
                The current HTTP request (used to check display preferences
                for the logged-in user).
        """
        self.review_request_details = review_request_details
        self.request = request

    @property
    def value(self):
        """Return the value loaded from the database.

        This will fetch the value with the associated ReviewRequest or
        ReviewRequestDraft, and then cache it for future lookups.

        Returns:
            object:
            The value of the field.
        """
        if not hasattr(self, '_value'):
            self._value = self.load_value(self.review_request_details)

        return self._value

    def has_value_changed(self, old_value, new_value):
        """Return whether the value has changed.

        By default, it performs an inequality check on the values. This
        can be overridden to perform more specialized checks.

        Args:
            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.

        Returns:
            bool:
            Whether the value of the field has changed.
        """
        return old_value != new_value

    def record_change_entry(self, changedesc, old_value, new_value):
        """Record information on the changed values in a ChangeDescription.

        By default, the values are stored as-is along with the field ID.
        This can be overridden to perform more specialized storage.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the entry in.

            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.
        """
        changedesc.record_field_change(self.field_id, old_value, new_value)

    def serialize_change_entry(self, changedesc):
        """Serialize a change entry for public consumption.

        This will output a version of the change entry for use in the API.
        It can be the same content stored in the
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`, but
        does not need to be.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description whose field is to be serialized.

        Returns:
            dict:
            An appropriate serialization for the field.
        """
        field_info = changedesc.fields_changed[self.field_id]

        if self.model:
            return self.serialize_change_entry_for_model_list(field_info)
        else:
            return self.serialize_change_entry_for_singleton(field_info)

    def serialize_change_entry_for_model_list(self, field_info):
        """Return the change entry for a list of models.

        Args:
            field_info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            dict:
            A mapping of each key present in ``field_info`` to its list of
            model instances.
        """
        pks = [
            value[2]
            for key in ('new', 'old', 'added', 'removed')
            if key in field_info
            for value in field_info[key]
        ]
        pk_to_objects = dict([
            (obj.pk, obj)
            for obj in self.model.objects.filter(pk__in=pks)
        ])

        return dict([
            (key, [
                pk_to_objects[value[2]]
                for value in field_info[key]
            ])
            for key in ('new', 'old', 'added', 'removed')
            if key in field_info
        ])

    def serialize_change_entry_for_singleton(self, field_info):
        """Return the change entry for a singleton.

        Singleton fields (e.g., summaries) are stored in
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`\s as
        a list with a single element.

        Args:
            field_info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            dict:
            A mapping of each key in ``field_info`` to a single value.
        """
        return dict([
            (key, field_info[key][0])
            for key in ('new', 'old', 'added', 'removed')
            if key in field_info
        ])

    def serialize_change_entry_for_list(self, field_info):
        """Return the change entry for a list of plain data.

        Args:
            field_info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            dict:
            A mapping of each key in ``field_info`` to a list of values.
        """
        return dict([
            (key, [
                value[0]
                for value in field_info[key]
            ])
            for key in ('new', 'old', 'added', 'removed')
            if key in field_info
        ])

    def get_change_entry_sections_html(self, info):
        """Return sections of change entries with titles and rendered HTML.

        By default, this just returns a single section for the field, with
        the field's title and rendered change HTML.

        Subclasses can override this to provide more information.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            list of dict:
            A list of the change entry sections.
        """
        return [{
            'title': self.label,
            'rendered_html': mark_safe(self.render_change_entry_html(info)),
        }]

    def render_change_entry_html(self, info):
        """Render a change entry to HTML.

        By default, this returns a simple "changed from X to Y" using the old
        and new values. This can be overridden to generate more specialized
        output.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Subclasses can override ``render_change_entry_value_html`` to
        change how the value itself will be rendered in the string.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            unicode:
            The HTML representation of the change entry.
        """
        old_value = ''
        new_value = ''

        if 'old' in info:
            old_value = info['old'][0]

        if 'new' in info:
            new_value = info['new'][0]

        s = ['<table class="changed">']

        if old_value:
            s.append(self.render_change_entry_removed_value_html(
                info, old_value))

        if new_value:
            s.append(self.render_change_entry_added_value_html(
                info, new_value))

        s.append('</table>')

        return ''.join(s)

    def render_change_entry_added_value_html(self, info, value):
        """Render the change entry for an added value to HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            value (object):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        value_html = self.render_change_entry_value_html(info, value)

        if value_html:
            return ('<tr class="new-value"><th class="marker">+</th>'
                    '<td class="value">%s</td></tr>' % value_html)
        else:
            return ''

    def render_change_entry_removed_value_html(self, info, value):
        """Render the change entry for a removed value to HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            value (object):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        value_html = self.render_change_entry_value_html(info, value)

        if value_html:
            return ('<tr class="old-value"><th class="marker">-</th>'
                    '<td class="value">%s</td></tr>' % value_html)
        else:
            return ''

    def render_change_entry_value_html(self, info, value):
        """Render the value for a change description string to HTML.

        By default, this just converts the value to text and escapes it.
        This can be overridden to customize how the value is displayed.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            value (object):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        return escape(six.text_type(value or ''))

    def load_value(self, review_request_details):
        """Load a value from the review request or draft.

        By default, this loads the value as-is from the extra_data field.
        This can be overridden if you need to deserialize the value in some
        way.

        This must use ``review_request_details`` instead of
        ``self.review_request_details``.

        Args:
            review_request_details (reviewboard.reviews.models.base_review_request_details.BaseReviewRequestDetails):
                The review request or draft.
        """
        return review_request_details.extra_data.get(self.field_id)

    def save_value(self, value):
        """Save the value in the review request or draft.

        By default, this saves the value as-is in the extra_data field.
        This can be overridden if you need to serialize the value in some
        way.

        Args:
            value (object):
                The new value for the field.
        """
        self.review_request_details.extra_data[self.field_id] = value

    def propagate_data(self, review_request_details):
        """Propagate data in from source review request or draft.

        By default, this loads only the field's value from a source review
        request or draft and saves it as-is into the review request or draft
        associated with the field. This can be overridden if you need to
        propagate additional data elements.

        This method is preferable to explictly calling :py:meth:`load_value`
        and :py:meth:`save_value` in series to propagate data from a source
        into a field, because it allows for copying additional data elements
        beyond only the field's value.

        This function must use the ``review_request_details`` parameter instead
        of the :py:attr:`review_request_details` attribute on the field.

        Args:
            review_request_details (reviewboard.reviews.models.base_review_request_details.BaseReviewRequestDetails):
                The source review request or draft whose data is to be
                propagated.
        """
        self.save_value(self.load_value(review_request_details))

    def render_value(self, value):
        """Render the value in the field.

        By default, this converts to text and escapes it. This can be
        overridden if you need to render it in a more specific way.

        This must use ``value`` instead of ``self.value``.

        Args:
            value (object):
                The value to render.

        Returns:
            unicode:
            The rendered value.
        """
        return escape(six.text_type(value or ''))

    def get_css_classes(self):
        """Return the set of CSS classes to apply to the element.

        By default, this will include the contents of ``default_css_classes``,
        and ``required`` if it's a required field.

        This can be overridden to provide additional CSS classes, if they're
        not appropraite for ``default_css_classes``.

        Returns:
            set of unicode:
            A set of the CSS classes to apply.
        """
        css_classes = set(self.default_css_classes)

        if self.is_required:
            css_classes.add('required')

        return css_classes

    def get_dom_attributes(self):
        """Return any additional attributes to include in the element.

        By default, this returns nothing.

        Returns:
            dict:
            Additional key/value pairs for attributes to include in the
            rendered HTML element.
        """
        return {}

    def get_data_attributes(self):
        """Return any data attributes to include in the element.

        By default, this returns nothing.

        Returns:
            dict:
            The data attributes to include in the element.
        """
        return {}

    def as_html(self):
        """Return the field rendered to HTML.

        Returns:
            django.utils.safetext.SafeString:
            The rendered field.
        """
        return render_to_string('reviews/review_request_field.html', {
            'field': self,
        })

    def value_as_html(self):
        """Return the field rendered as HTML.

        By default, this just calls ``render_value`` with the value
        from the database.

        Returns:
            unicode:
            The rendered field.
        """
        return self.render_value(self.value)

    def __str__(self):
        """Represent the field as a byte string.

        Returns:
            bytes:
            The field's ID as a byte string.
        """
        if isinstance(self.field_id, six.binary_type):
            return self.field_id

        return self.field_id.encode('utf-8')

    def __unicode__(self):
        """Represent the field as a unicode string.

        Returns:
            unicode:
            The field's ID as a unicode string.
        """
        if isinstance(self.field_id, six.binary_type):
            return self.field_id.decode('utf-8')

        return self.field_id


class BaseEditableField(BaseReviewRequestField):
    """Base class for an editable field.

    This simply marks the field as editable.
    """

    default_css_classes = ['editable']
    is_editable = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.TextFieldView'


class BaseCommaEditableField(BaseEditableField):
    """Base class for an editable comma-separated list of values.

    This is used for dealing with lists of items that appear
    comma-separated in the UI. It works with stored lists of content
    on the review request or draft, and on the ChangeDescription.

    Subclasses can override this to provide specialized rendering
    on a per-item-basis. That's useful for showing links to items,
    for example.
    """

    default_css_classes = ['editable', 'comma-editable']
    order_matters = False

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.CommaSeparatedValuesTextFieldView'

    one_line_per_change_entry = True

    def has_value_changed(self, old_value, new_value):
        """Return whether two values have changed.

        If :py:attr:`order_matters` is set to ``True``, this will do a strict
        list comparison. Otherwise, it will compare the items in both lists
        without caring about the ordering.

        Args:
            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.

        Returns:
            bool:
            Whether the value of the field has changed.
        """
        if self.order_matters:
            return old_value != new_value
        else:
            return set(old_value or []) != set(new_value or [])

    def serialize_change_entry(self, changedesc):
        """Serialize a change entry for public consumption.

        This will output a version of the change entry for use in the API.
        It can be the same content stored in the
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`, but
        does not need to be.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description whose field is to be serialized.

        Returns:
            dict:
            An appropriate serialization for the field.
        """
        field_info = changedesc.fields_changed[self.field_id]

        if self.model:
            return self.serialize_change_entry_for_model_list(field_info)
        else:
            return self.serialize_change_entry_for_list(field_info)

    def render_value(self, values):
        """Render the list of items.

        This will call out to ``render_item`` for every item. The list
        of rendered items will be separated by a comma and a space.

        Args:
            value (object):
                The value to render.

        Returns:
            unicode:
            The rendered value.
        """
        return ', '.join([
            self.render_item(value)
            for value in values
        ])

    def render_item(self, item):
        """Render an item from the list.

        By default, this will convert the item to text and then escape it.

        Args:
            item (object):
                The item to render.

        Returns:
            unicode:
            The rendered item.
        """
        return escape(six.text_type(item or ''))

    def render_change_entry_html(self, info):
        """Render a change entry to HTML.

        By default, this returns HTML containing a list of removed items,
        and a list of added items. This can be overridden to generate
        more specialized output.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            unicode:
            The HTML representation of the change entry.
        """
        s = ['<table class="changed">']

        if 'removed' in info:
            values = info['removed']

            if self.one_line_per_change_entry:
                s += [
                    self.render_change_entry_removed_value_html(info, [value])
                    for value in values
                ]
            else:
                s.append(self.render_change_entry_removed_value_html(
                    info, values))

        if 'added' in info:
            values = info['added']

            if self.one_line_per_change_entry:
                s += [
                    self.render_change_entry_added_value_html(info, [value])
                    for value in values
                ]
            else:
                s.append(self.render_change_entry_added_value_html(
                    info, values))

        s.append('</table>')

        return ''.join(s)

    def render_change_entry_value_html(self, info, values):
        """Render a list of items for change description HTML.

        By default, this will call ``render_change_entry_item_html`` for every
        item in the list. The list of rendered items will be separated by a
        comma and a space.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            values (list):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        return ', '.join([
            self.render_change_entry_item_html(info, item)
            for item in values
        ])

    def render_change_entry_item_html(self, info, item):
        """Render an item for change description HTML.

        By default, this just converts the value to text and escapes it.
        This can be overridden to customize how the value is displayed.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the item.

        Returns:
            unicode:
            The rendered change entry.
        """
        return escape(six.text_type(item[0]))


class BaseTextAreaField(BaseEditableField):
    """Base class for a multi-line text area field.

    The text area can take either plain text or Markdown text. By default,
    Markdown is supported, but this can be changed by setting
    ``enable_markdown`` to ``False``.
    """

    default_css_classes = ['editable', 'field-text-area']
    enable_markdown = True
    always_render_markdown = False
    tag_name = 'pre'

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.MultilineTextFieldView'

    @cached_property
    def text_type_key(self):
        """Return the text type key for the ``extra_data`` dictionary.

        Returns:
            unicode:
            The key which stores the text type for the field.
        """
        if self.field_id == 'text':
            return 'text_type'
        else:
            return '%s_text_type' % self.field_id

    def is_text_markdown(self, value):
        """Return whether the text is in Markdown format.

        This can be overridden if the field needs to check something else
        to determine if the text is in Markdown format.

        Args:
            value (unicode):
                The value of the field.

        Returns:
            bool:
            Whether the text should be interpreted as Markdown.
        """
        text_type = self.review_request_details.extra_data.get(
            self.text_type_key, 'plain')

        return text_type == 'markdown'

    def propagate_data(self, review_request_details):
        """Propagate data in from source review request or draft.

        In addition to the value propagation handled by the base class, this
        copies the text type details from a source review request or draft and
        saves it as-is into the review request or draft associated with the
        field.

        Args:
            review_request_details (reviewboard.reviews.models.base_review_request_details.BaseReviewRequestDetails):
                The source review request or draft whose data is to be
                propagated.
        """
        super(BaseTextAreaField, self).propagate_data(review_request_details)

        source_text_type = review_request_details.extra_data.get(
            self.text_type_key, None)

        if source_text_type is not None:
            self.review_request_details.extra_data[self.text_type_key] = \
                source_text_type

    def get_css_classes(self):
        """Return the list of CSS classes.

        If Markdown is enabled, and the text is in Markdown format,
        this will add a "rich-text" field.

        Returns:
            set of unicode:
            A set of the CSS classes to apply.
        """
        css_classes = super(BaseTextAreaField, self).get_css_classes()

        if (self.enable_markdown and self.value and
            (self.should_render_as_markdown(self.value) or
             (self.request.user and
              is_rich_text_default_for_user(self.request.user)))):
            css_classes.add('rich-text')

        return css_classes

    def get_data_attributes(self):
        """Return any data attributes to include in the element.

        By default, this returns nothing.

        Returns:
            dict:
            The data attributes to include in the element.
        """
        attrs = super(BaseTextAreaField, self).get_data_attributes()

        if self.enable_markdown:
            if self.request:
                user = self.request.user
            else:
                user = None

            attrs.update({
                'allow-markdown': True,
                'raw-value': normalize_text_for_edit(
                    user, self.value,
                    self.should_render_as_markdown(self.value)),
            })

        return attrs

    def render_value(self, text):
        """Return the value of the field.

        If Markdown is enabled, and the text is not in Markdown format,
        the text will be escaped.

        Args:
            text (unicode):
                The value of the field.

        Returns:
            unicode:
            The rendered value.
        """
        text = text or ''

        if self.should_render_as_markdown(text):
            return render_markdown(text)
        else:
            return escape(text)

    def should_render_as_markdown(self, value):
        """Return whether the text should be rendered as Markdown.

        By default, this checks if the field is set to always render
        any text as Markdown, or if the given text is in Markdown format.

        Args:
            value (unicode):
                The value of the field.

        Returns:
            bool:
            Whether the text should be rendered as Markdown.
        """
        return self.always_render_markdown or self.is_text_markdown(value)

    def render_change_entry_html(self, info):
        """Render a change entry to HTML.

        This will render a diff of the changed text.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            unicode:
            The HTML representation of the change entry.
        """
        old_value = ''
        new_value = ''

        if 'old' in info:
            old_value = info['old'][0] or ''

        if 'new' in info:
            new_value = info['new'][0] or ''

        old_value = render_markdown(old_value)
        new_value = render_markdown(new_value)
        old_lines = list(iter_markdown_lines(old_value))
        new_lines = list(iter_markdown_lines(new_value))

        differ = MyersDiffer(old_lines, new_lines)

        return ('<table class="diffed-text-area">%s</table>'
                % ''.join(self._render_all_change_lines(differ, old_lines,
                                                        new_lines)))

    def _render_all_change_lines(self, differ, old_lines, new_lines):
        for tag, i1, i2, j1, j2 in differ.get_opcodes():
            if tag == 'equal':
                lines = self._render_change_lines(differ, tag, None, None,
                                                  i1, i2, old_lines)
            elif tag == 'insert':
                lines = self._render_change_lines(differ, tag, None, '+',
                                                  j1, j2, new_lines)
            elif tag == 'delete':
                lines = self._render_change_lines(differ, tag, '-', None,
                                                  i1, i2, old_lines)
            elif tag == 'replace':
                lines = self._render_change_replace_lines(differ, i1, i2,
                                                          j1, j2, old_lines,
                                                          new_lines)
            else:
                raise ValueError('Unexpected tag "%s"' % tag)

            for line in lines:
                yield line

    def _render_change_lines(self, differ, tag, old_marker, new_marker,
                             i1, i2, lines):
        old_marker = old_marker or '&nbsp;'
        new_marker = new_marker or '&nbsp;'

        for i in range(i1, i2):
            line = lines[i]

            yield ('<tr class="%s">'
                   ' <td class="marker">%s</td>'
                   ' <td class="marker">%s</td>'
                   ' <td class="line rich-text">%s</td>'
                   '</tr>'
                   % (tag, old_marker, new_marker, line))

    def _render_change_replace_lines(self, differ, i1, i2, j1, j2,
                                     old_lines, new_lines):
        replace_new_lines = []

        for i, j in zip(range(i1, i2), range(j1, j2)):
            old_line = old_lines[i]
            new_line = new_lines[j]

            parser = HTMLParser()

            old_regions, new_regions = \
                get_line_changed_regions(parser.unescape(strip_tags(old_line)),
                                         parser.unescape(strip_tags(new_line)))

            old_line = highlightregion(old_line, old_regions)
            new_line = highlightregion(new_line, new_regions)

            yield (
                '<tr class="replace-old">'
                ' <td class="marker">~</td>'
                ' <td class="marker">&nbsp;</td>'
                ' <td class="line rich-text">%s</td>'
                '</tr>'
                % old_line)

            replace_new_lines.append(new_line)

        for line in replace_new_lines:
            yield (
                '<tr class="replace-new">'
                ' <td class="marker">&nbsp;</td>'
                ' <td class="marker">~</td>'
                ' <td class="line rich-text">%s</td>'
                '</tr>'
                % line)


class BaseCheckboxField(BaseReviewRequestField):
    """Base class for a checkbox.

    The field's value will be either True or False.
    """

    is_editable = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.CheckboxFieldView'

    #: The default value of the field.
    default_value = False

    #: The HTML tag to be used when rendering the field.
    tag_name = 'input'

    def load_value(self, review_request_details):
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            bool:
            The loaded value.
        """
        value = review_request_details.extra_data.get(self.field_id)

        if value is not None:
            return value
        else:
            return self.default_value

    def render_change_entry_html(self, info):
        """Render a change entry to HTML.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            unicode:
            The HTML representation of the change entry.
        """
        old_value = None
        new_value = None

        if 'old' in info:
            old_value = info['old'][0]

        if 'new' in info:
            new_value = info['new'][0]

        s = ['<table class="changed">']

        if old_value is not None:
            s.append(self.render_change_entry_removed_value_html(
                info, old_value))

        if new_value is not None:
            s.append(self.render_change_entry_added_value_html(
                info, new_value))

        s.append('</table>')

        return ''.join(s)

    def render_change_entry_value_html(self, info, value):
        """Render the value for a change description string to HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        if value:
            checked = 'checked'
        else:
            checked = ''

        return ('<input type="checkbox" autocomplete="off" disabled %s>'
                % checked)

    def get_dom_attributes(self):
        """Return any additional attributes to include in the element.

        By default, this returns nothing.

        Returns:
            dict:
            Additional key/value pairs for attributes to include in the
            rendered HTML element.
        """
        attrs = {
            'type': 'checkbox',
        }

        if self.value:
            attrs['checked'] = 'checked'

        return attrs

    def value_as_html(self):
        """Return the field rendered as HTML.

        Because the value is included as a boolean attribute on the checkbox
        element, this just returns the empty string.

        Returns:
            unicode:
            The rendered field.
        """
        return ''


class BaseDropdownField(BaseReviewRequestField):
    """Base class for a drop-down field."""

    is_editable = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.DropdownFieldView'

    #: The default value of the field.
    default_value = None

    #: The HTML tag to be used when rendering the field.
    tag_name = 'select'

    #: A list of the available options for the dropdown.
    #:
    #: Each entry in the list should be a 2-tuple of (value, label). The values
    #: must be unique. Both values and labels should be unicode.
    options = []

    def load_value(self, review_request_details):
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            unicode:
            The loaded value.
        """
        value = review_request_details.extra_data.get(self.field_id)

        if value is not None:
            return value
        else:
            return self.default_value

    def render_change_entry_value_html(self, info, value):
        """Render the value for a change description string to HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the field.

        Returns:
            unicode:
            The rendered change entry.
        """
        for key, label in self.options:
            if value == key:
                return escape(label)

        return ''

    def value_as_html(self):
        """Return the field rendered as HTML.

        Select tags are funny kinds of inputs, and need a bunch of
        ``<option>`` elements inside them. This renders the "value" of the
        field as those options, to fit in with the base field's template.

        Returns:
            django.utils.safestring.SafeText:
            The rendered field.
        """
        data = []

        for value, label in self.options:
            if self.value == value:
                selected = ' selected'
            else:
                selected = ''

            data.append((value, selected, label))

        return format_html_join(
            '',
            '<option value="{}"{}>{}</option>',
            data)


class BaseDateField(BaseEditableField):
    """Base class for a date field."""

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.DateFieldView'

    #: The default value of the field.
    default_value = ''

    def load_value(self, review_request_details):
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            unicode:
            The loaded value.
        """
        value = review_request_details.extra_data.get(self.field_id)

        if value is not None:
            return value
        else:
            return self.default_value


def get_review_request_fields():
    """Yield all registered field classes.

    Yields:
        type:
        The field classes, as subclasses of :py:class:`BaseReviewRequestField`
    """
    for field in field_registry:
        yield field


def get_review_request_fieldsets(include_change_entries_only=False):
    """Return a list of all registered fieldset classes.

    As an internal optimization, the "main" fieldset can be filtered out,
    to help with rendering the side of the review request page.

    Args:
        include_change_entries_only (bool):
            Whether or not to include the change-entry only fieldset.

    Returns:
        list:
        The requested :py:class:`fieldsets <BaseReviewRequestFieldSet>`.
    """
    excluded_ids = []

    if not include_change_entries_only:
        excluded_ids.append('_change_entries_only')

    return [
        fieldset
        for fieldset in fieldset_registry
        if fieldset.fieldset_id not in excluded_ids
    ]


def get_review_request_fieldset(fieldset_id):
    """Return the fieldset with the specified ID.

    Args:
        fieldset_id (unicode):
            The fieldset's ID.

    Returns:
        BaseReviewRequestFieldSet:
        The requested fieldset, or ``None`` if it could not be found.
    """
    return fieldset_registry.get('fieldset_id', fieldset_id)


def get_review_request_field(field_id):
    """Return the field with the specified ID.

    Args:
        field_id (unicode):
            The field's ID.

    Returns:
        BaseReviewRequestField:
        The requested field, or ``None`` if it could not be found.
    """
    return field_registry.get('field_id', field_id)


def register_review_request_fieldset(fieldset):
    """Register a custom review request fieldset.

    The fieldset must have a :py:attr:`~BaseReviewRequestFieldSet.fieldset_id`
    attribute. This ID **must** be unique across all registered fieldsets, or
    an exception will be thrown.

    Args:
        fieldset (type):
            The :py:class:`BaseReviewRequestFieldSet` subclass.

    Raises:
        djblets.registries.errors.ItemLookupError:
            This will be thrown if a fieldset is already registered with the
            same ID.
    """
    fieldset_registry.register(fieldset)


def unregister_review_request_fieldset(fieldset):
    """Unregister a previously registered review request fieldset.

    Args:
        fieldset (type):
            The :py:class:`BaseReviewRequestFieldSet` subclass.

    Raises:
        djblets.registries.errors.ItemLookupError:
            This will be thrown if the fieldset is not already registered.
    """
    try:
        fieldset_registry.unregister(fieldset)
    except ItemLookupError as e:
        logging.error('Failed to unregister unknown review request fieldset '
                      '"%s"',
                      fieldset.fieldset_id)
        raise e



from __future__ import unicode_literals

from django.utils import six
from djblets.db.query import LocalDataQuerySet
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.responses import WebAPIResponsePaginated

from reviewboard.hostingsvcs.repository import RemoteRepository
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_local_site,
                                           webapi_check_login_required)
from reviewboard.webapi.resources import resources


class RemoteRepositoryResponsePaginated(WebAPIResponsePaginated):
    """Provides paginated reponses for lists of RemoteRepository objects.

    This is a specialization of WebAPIResponsePaginated designed to
    return lsits of RemoteRepository objects and to handle pagination in
    a way that's compatible with the pagination models of HostingService.
    """
    def __init__(self, request, queryset, *args, **kwargs):
        self.paginator = queryset[0]

        super(RemoteRepositoryResponsePaginated, self).__init__(
            request, queryset=None, *args, **kwargs)

    def has_prev(self):
        return self.paginator.has_prev

    def has_next(self):
        return self.paginator.has_next

    def get_prev_index(self):
        return max(self.start - 1, 0)

    def get_next_index(self):
        return self.start + 1

    def get_results(self):
        return self.paginator.page_data

    def get_total_results(self):
        return None

    def build_pagination_url(self, full_path, start, max_results,
                             query_parameters):
        return '%s?start=%s%s' % (full_path, start, query_parameters)


class RemoteRepositoryResource(WebAPIResource):
    """Returns information on remote repositories on a hosting service.

    This can be used to look up the information needed to connect to a
    remote repository or to add a repository to Review Board. Only remote
    repositories that are accessible to the linked hosting service account
    (i.e., that of the parent resource) will be provided by this resource.
    """
    added_in = '2.5'

    name = 'remote_repository'
    name_plural = 'remote_repositories'

    model = RemoteRepository
    model_object_key = 'id'
    model_parent_key = 'hosting_service_account'
    uri_object_key = 'repository_id'
    uri_object_key_regex = r'[A-Za-z0-9_./-]+'

    paginated_cls = RemoteRepositoryResponsePaginated

    fields = {
        'id': {
            'type': six.text_type,
            'description': 'The unique ID for this repository on the '
                           'hosting service.',
        },
        'name': {
            'type': six.text_type,
            'description': 'The name of the repository.',
        },
        'owner': {
            'type': six.text_type,
            'description': 'The owner of the repository, which may be a user '
                           'account or an organization, depending on the '
                           'service.',
        },
        'scm_type': {
            'type': six.text_type,
            'description': 'The type of repository, mapping to registered '
                           'SCMTools on Review Board.',
        },
        'path': {
            'type': six.text_type,
            'description': 'The repository path as recommended by the hosting '
                           'service.',
        },
        'mirror_path': {
            'type': six.text_type,
            'description': 'A secondary path that can be used to reach the '
                           'repository.',
        },
    }
    uri_object_key = 'repository_id'
    autogenerate_etags = True

    allowed_methods = ('GET',)

    def has_list_access_permissions(self, request, *args, **kwargs):
        account = resources.hosting_service_account.get_object(
            request, *args, **kwargs)

        # Only allow administrators or those with the ability to modify the
        # account to see what repositories are listed.
        return account.is_mutable_by(request.user)

    def has_access_permissions(self, request, remote_repository,
                               *args, **kwargs):
        # Only allow administrators or those with the ability to modify the
        # account to see what repositories are listed.
        return remote_repository.hosting_service_account.is_mutable_by(
            request.user)

    def get_queryset(self, request, start=None, is_list=False,
                     repository_id=None, *args, **kwargs):
        account = resources.hosting_service_account.get_object(
            request, *args, **kwargs)

        if is_list:
            # Wrap the paginator in a LocalDataQuerySet, so that we can get
            # to it later in RemoteRepositoryResponsePaginated.
            lookup_kwargs = {}

            for name in ('owner', 'owner-type', 'filter-type'):
                if kwargs.get(name):
                    arg = name.replace('-', '_')
                    lookup_kwargs[arg] = kwargs[name]

            result = account.service.get_remote_repositories(start=start,
                                                             **lookup_kwargs)
        else:
            result = account.service.get_remote_repository(repository_id)

        return LocalDataQuerySet([result])

    def get_serializer_for_object(self, obj):
        if isinstance(obj, RemoteRepository):
            return self

        return super(RemoteRepositoryResource, self).get_serializer_for_object(
            obj)

    # NOTE: We're not augmenting from any resources, because we don't want to
    #       include ?counts-only= or ?max-results=, and we have a different
    #       ?start=. Because of this, we need to be careful to apply our own
    #       decorators.
    @webapi_check_login_required
    @webapi_check_local_site
    @webapi_request_fields(
        optional={
            'owner': {
                'type': six.text_type,
                'description': 'The owner (user account or organization) '
                               'to look up repositories for. Defaults to '
                               'the owner of the hosting service account.',
            },
            'owner-type': {
                'type': six.text_type,
                'description': 'Indicates what sort of account the owner '
                               'represents. This may be required by some '
                               'services, and the values are dependent on '
                               'that service.',
            },
            'filter-type': {
                'type': six.text_type,
                'description': 'Filters the list of results. Allowed values '
                               'are dependent on the hosting service. '
                               'Unexpected values will be ignored.',
            },
            'start': {
                'type': int,
                'description': 'The 0-based index of the first page of '
                               'results to fetch.',
            },
        },
        allow_unknown=True
    )
    def get_list(self, request, *args, **kwargs):
        """Returns the list of remote repositories on the hosting service.

        Different hosting service backends have different criteria for
        performing the lookups. Some hosting services have multiple types of
        owners, specified by passing ``owner-type``. Filtering may also be
        possible by passing ``filter-type``. Performing lookups requires
        knowing the possible values for the service ahead of time and passing
        the proper parameters in the query string.

        Pagination works a bit differently for this resource than most.
        Instead of ``?start=`` taking an index into the number of results,
        this resource's ``?start=`` takes a 0-based index of the page of
        results.

        ``?max-results=`` and ``?counts-only=`` are not supported, as they're
        not compatible with all hosting services.

        Callers should always use the ``next`` and ``prev`` links for
        navigation, and should not build page indexes themselves.
        """
        return super(RemoteRepositoryResource, self).get_list(
            request, *args, **kwargs)

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Provides information on a particular remote repository.

        If the remote repository exists and is accessible by the linked
        hosting service account (that of the parent resource), then the
        details of that repository will be returned in the payload.

        The ID expected for the lookup in the URL is specific to the type
        of hosting service.
        """
        pass


remote_repository_resource = RemoteRepositoryResource()



"""Unit tests for reviewboard.reviews.views.DownloadDiffFileView."""

from __future__ import unicode_literals

from reviewboard.extensions.tests import TestService
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.service import (register_hosting_service,
                                             unregister_hosting_service)
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class DownloadDiffFileViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.DownloadDiffFileView."""

    fixtures = ['test_users', 'test_scmtools']

    @classmethod
    def setUpClass(cls):
        super(DownloadDiffFileViewTests, cls).setUpClass()

        register_hosting_service(TestService.name, TestService)

    @classmethod
    def tearDownClass(cls):
        super(DownloadDiffFileViewTests, cls).tearDownClass()

        unregister_hosting_service(TestService.name)

    def setUp(self):
        super(DownloadDiffFileViewTests, self).setUp()

        self.account = HostingServiceAccount.objects.create(
            service_name=TestService.name,
            hosting_url='http://example.com/',
            username='foo')

        self.repository = self.create_repository(hosting_account=self.account)
        self.review_request = self.create_review_request(
            repository=self.repository, publish=True)
        self.diffset = self.create_diffset(review_request=self.review_request)
        self.filediff = self.create_filediff(self.diffset,
                                             source_file='/invalid-path',
                                             dest_file='/invalid-path')

    def testing_download_orig_file_404(self):
        """Testing DownloadDiffFileView with original file when the file
        cannot be found upstream
        """
        rsp = self.client.get(
            local_site_reverse('download-orig-file', kwargs={
                'review_request_id': self.review_request.display_id,
                'revision': self.diffset.revision,
                'filediff_id': self.filediff.pk,
            }))

        self.assertEquals(rsp.status_code, 404)

    def testing_download_modified_file_404(self):
        """Testing DownloadDiffFileView with modified file when the file
        cannot be found upstream
        """
        rsp = self.client.get(
            local_site_reverse('download-modified-file', kwargs={
                'review_request_id': self.review_request.display_id,
                'revision': self.diffset.revision,
                'filediff_id': self.filediff.pk,
            }))

        self.assertEquals(rsp.status_code, 404)



"""The signal processor for Review Board search."""

from __future__ import unicode_literals

import threading
from functools import partial

from django.contrib.auth.models import User
from django.db.models.signals import post_delete, post_save, m2m_changed
from django.utils import six
from haystack.signals import BaseSignalProcessor

from reviewboard.accounts.models import Profile
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.reviews.signals import review_request_published
from reviewboard.search import search_backend_registry


class SignalProcessor(BaseSignalProcessor):
    """"Listens for signals and updates the search index.

    This will listen for any signals that would affect the search index, and
    invokes a suitable Haystack callback to immediately update the data stored
    in the index.

    This only updates the search index if:

    1) Search is enabled.
    2) The current search engine backend supports on-the-fly indexing.
    """

    save_signals = [
        (ReviewRequest, review_request_published, 'review_request'),
        (User, post_save, 'instance'),
        (Profile, post_save, 'instance'),
    ]

    delete_signals = [
        (ReviewRequest, post_delete),
        (User, post_delete),
    ]

    def __init__(self, *args, **kwargs):
        """Initialize the signal processor.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent constructor.

            **kwargs (dict):
                Keyword arguments to pass to the parent constructor.
        """
        self.is_setup = False
        self._handlers = {}
        self._pending_user_changes = threading.local()

        super(SignalProcessor, self).__init__(*args, **kwargs)

    def setup(self):
        """Register the signal handlers for this processor."""

        # We define this here instead of at the class level because we cannot
        # reference class members during the class' definition.
        m2m_changed_signals = [
            (Group.users.through, self._handle_group_m2m_changed),
        ]

        if not self.is_setup:
            for cls, signal, instance_kwarg in self.save_signals:
                handler = partial(self.check_handle_save,
                                  instance_kwarg=instance_kwarg)
                self._handlers[(cls, signal)] = handler

            for cls, signal in self.delete_signals:
                self._handlers[(cls, signal)] = self.check_handle_delete

            for cls, handler in m2m_changed_signals:
                self._handlers[(cls, m2m_changed)] = handler

            for (cls, signal), handler in six.iteritems(self._handlers):
                signal.connect(handler, sender=cls)

            self.is_setup = True

    def teardown(self):
        """Unregister all signal handlers for this processor."""
        if self.is_setup:
            for (cls, signal), handler in six.iteritems(self._handlers):
                signal.disconnect(handler, sender=cls)

            self.is_setup = False

    def check_handle_save(self, instance_kwarg, **kwargs):
        """Conditionally update the search index when an object is updated.

        Args:
            instance_kwarg (unicode):
                The name of the instance parameter.

            **kwargs (dict):
                Signal arguments. These will be passed to
                :py:meth:`handle_save`.
        """
        instance = kwargs.pop(instance_kwarg)
        backend = search_backend_registry.current_backend

        if backend and search_backend_registry.on_the_fly_indexing_enabled:
            if isinstance(instance, Profile):
                # When we save a Profile, we want to update the User index.
                kwargs['sender'] = User
                instance = instance.user

            self.handle_save(instance=instance, **kwargs)

    def check_handle_delete(self, **kwargs):
        """Conditionally update the search index when an object is deleted.

        Args:
            **kwargs (dict):
                Signal arguments. These will be passed to
                :py:meth:`handle_delete`.
        """
        backend = search_backend_registry.current_backend

        if backend and search_backend_registry.on_the_fly_indexing_enabled:
            self.handle_delete(**kwargs)

    def _handle_group_m2m_changed(self, instance, action, pk_set, reverse,
                                  **kwargs):
        """Handle a Group.users relation changing.

        When the :py:attr:`Group.users
        <reviewboard.reviews.models.group.Group.users>` field changes, we don't
        get a corresponding :py:data:`~django.db.signals.post_save` signal
        (because the related model wasn't saved). Instead, we will get multiple
        :py:data:`~django.db.signals.m2m_changed` signals that indicate how the
        relation is changing. This method will handle those signals and
        call the correct save method so  that they can be re-indexed.

        Args:
            instance (django.contrib.auth.models.User or reviewboward.reviews.models.group.Group):
                The model that updated.

            action (unicode):
                The update action. This will be one of:

                * ``'pre_add'``
                * ``'post_add'``
                * ``'pre_remove'``
                * ``'post_remove'``
                * ``'pre_clear'``
                * ``'post_clear'``

            pk_set (set of int):
                The primary keys of the related objects that changed.

                When the action is ``'pre_clear'`` or ``'post_clear'``,
                this argument will be an empty set.

            reverse (bool):
                Whether or not the reverse relation was modified. If
                true, this indicated that ``instance`` is a
                :py:class:`~django.contrib.auth.models.User` object and
                ``pk_set`` is the set of primary keys of the added or removed
                groups.

                When this argument is false, ``instance`` is a
                :py:class:`~reviewboard.reviews.models.group.Group`
                object and ``pk_set`` is the set of primary keys of the added
                or removed users.

            **kwargs (dict):
                Additional keyword arguments.
        """
        backend = search_backend_registry.current_backend

        if not (backend and
                search_backend_registry.on_the_fly_indexing_enabled):
            return

        if not hasattr(self._pending_user_changes, 'data'):
            self._pending_user_changes.data = {}

        if action in ('post_add', 'post_remove'):
            if reverse:
                # When using the reverse relation, the instance is the User and
                # the pk_set is the PKs of the groups being added or removed.
                users = [instance]
            else:
                # Otherwise the instance is the Group and the pk_set is the set
                # of User primary keys.
                users = User.objects.filter(pk__in=pk_set)

            for user in users:
                self.handle_save(instance=user, instance_kwarg='instance',
                                 sender=User)
        elif action == 'pre_clear':
            # When ``reverse`` is ``True``, a User is having their groups
            # cleared so we don't need to worry about storing any state in the
            # pre_clear  phase.
            #
            # Otherwise, a ReviewGroup is having their users cleared. In both
            # the pre_clear and post_clear phases, the ``pk_set`` argument will
            # be empty, so we cache the PKs of the current members of the
            # groups so we know to reindex them.
            if not reverse:
                self._pending_user_changes.data[instance.pk] = list(
                    instance.users.values_list('pk', flat=True))
        elif action == 'post_clear':
            if reverse:
                # When ``reverse`` is ``True``, we just have to reindex a
                # single user.
                self.handle_save(instance=instance, instance_kwarg='instance',
                                 sender=User)
            else:
                # Here, we are reindexing every user that got removed from the
                # group via clearing.
                pks = self._pending_user_changes.data.pop(instance.pk)

                for user in User.objects.filter(pk__in=pks):
                    self.handle_save(instance=user, instance_kwarg='instance',
                                     sender=User)



"""Unit tests for the SourceForge hosting service."""

from __future__ import unicode_literals

from reviewboard.hostingsvcs.tests.testcases import ServiceTests


class SourceForgeTests(ServiceTests):
    """Unit tests for the SourceForge hosting service."""

    service_name = 'sourceforge'

    def test_service_support(self):
        """Testing SourceForge service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)

    def test_get_repository_fields_with_bazaar(self):
        """Testing SourceForge.get_repository_fields for Bazaar"""
        self.assertEqual(
            self.get_repository_fields(
                'Bazaar',
                fields={
                    'sourceforge_project_name': 'myproj',
                }
            ),
            {
                'path': 'bzr://myproj.bzr.sourceforge.net/bzrroot/myproj',
                'mirror_path': ('bzr+ssh://myproj.bzr.sourceforge.net/bzrroot/'
                                'myproj'),
            })

    def test_get_repository_fields_with_cvs(self):
        """Testing SourceForge.get_repository_fields for CVS"""
        self.assertEqual(
            self.get_repository_fields(
                'CVS',
                fields={
                    'sourceforge_project_name': 'myproj',
                }
            ),
            {
                'path': (':pserver:anonymous@myproj.cvs.sourceforge.net:'
                         '/cvsroot/myproj'),
                'mirror_path': 'myproj.cvs.sourceforge.net/cvsroot/myproj',
            })

    def test_get_repository_fields_with_mercurial(self):
        """Testing SourceForge.get_repository_fields for Mercurial"""
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'sourceforge_project_name': 'myproj',
                }
            ),
            {
                'path': 'http://myproj.hg.sourceforge.net:8000/hgroot/myproj',
                'mirror_path': 'ssh://myproj.hg.sourceforge.net/hgroot/myproj',
            })

    def test_get_repository_fields_with_svn(self):
        """Testing SourceForge.get_repository_fields for Subversion"""
        self.assertEqual(
            self.get_repository_fields(
                'Subversion',
                fields={
                    'sourceforge_project_name': 'myproj',
                }
            ),
            {
                'path': 'http://myproj.svn.sourceforge.net/svnroot/myproj',
                'mirror_path': ('https://myproj.svn.sourceforge.net/svnroot/'
                                'myproj'),
            })



from __future__ import unicode_literals

import logging

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.forms.util import ErrorDict
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render_to_response
from django.utils import six
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import TemplateView
from djblets.auth.views import register
from djblets.configforms.views import ConfigPagesView
from djblets.features.decorators import feature_required
from djblets.forms.fieldsets import filter_fieldsets
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import augment_method_from
from djblets.views.generic.etag import ETagViewMixin

from reviewboard.accounts.backends import get_enabled_auth_backends
from reviewboard.accounts.forms.registration import RegistrationForm
from reviewboard.accounts.mixins import CheckLoginRequiredViewMixin
from reviewboard.accounts.models import Profile
from reviewboard.accounts.pages import AccountPage, OAuth2Page, PrivacyPage
from reviewboard.accounts.privacy import is_consent_missing
from reviewboard.avatars import avatar_services
from reviewboard.notifications.email.decorators import preview_email
from reviewboard.notifications.email.message import \
    prepare_password_changed_mail
from reviewboard.oauth.features import oauth2_service_feature
from reviewboard.oauth.forms import (UserApplicationChangeForm,
                                     UserApplicationCreationForm)
from reviewboard.oauth.models import Application
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin
from reviewboard.site.urlresolvers import local_site_reverse


class UserInfoboxView(CheckLoginRequiredViewMixin,
                      CheckLocalSiteAccessViewMixin,
                      ETagViewMixin,
                      TemplateView):
    """Displays information on a user, for use in user pop-up infoboxes.

    This is meant to be embedded in other pages, rather than being
    a standalone page.
    """

    template_name = 'accounts/user_infobox.html'

    def __init__(self, **kwargs):
        """Initialize a view for the request.

        Args:
            **kwargs (dict):
                Keyword arguments passed to :py:meth:`as_view`.
        """
        super(UserInfoboxView, self).__init__(**kwargs)

        self._lookup_user = None
        self._show_profile = None
        self._timezone = None

    def get_etag_data(self, request, username, *args, **kwargs):
        """Return an ETag for the view.

        This will look up some state needed for the request and generate a
        suitable ETag.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            username (unicode):
                The username of the user being looked up.

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (tuple):
                Keyword arguments to pass to the handler.

                These will be arguments provided by the URL pattern.

        Returns:
            unicode:
            The ETag for the page.
        """
        from reviewboard.extensions.hooks import UserInfoboxHook

        user = get_object_or_404(User, username=username)
        self._lookup_user = user

        try:
            profile = user.get_profile()
            self._show_profile = user.is_profile_visible(request.user)
            self._timezone = profile.timezone
        except Profile.DoesNotExist:
            self._show_profile = True
            self._timezone = 'UTC'

        etag_data = [
            user.first_name,
            user.last_name,
            user.email,
            six.text_type(user.last_login),
            six.text_type(settings.TEMPLATE_SERIAL),
            six.text_type(self._show_profile),
            self._timezone,
        ]

        if avatar_services.avatars_enabled:
            avatar_service = avatar_services.for_user(user)

            if avatar_service:
                etag_data.extend(avatar_service.get_etag_data(user))

        local_site = self.local_site

        for hook in UserInfoboxHook.hooks:
            try:
                etag_data.append(hook.get_etag_data(
                    user=user,
                    request=request,
                    local_site=local_site))
            except Exception as e:
                logging.exception('Error when running UserInfoboxHook.'
                                  'get_etag_data method in extension "%s": %s',
                                  hook.extension.id, e)

        return ':'.join(etag_data)

    def get_context_data(self, **kwargs):
        """Return data for the template.

        This will return information on the user, along with information from
        any extension hooks used for the page.

        Args:
            **kwargs (tuple):
                Additional keyword arguments from the URL pattern.

        Returns:
            dict:
            Context data for the template.
        """
        from reviewboard.extensions.hooks import UserInfoboxHook

        # These are accessed several times, so bring them in to reduce
        # attribute lookups.
        user = self._lookup_user
        username = user.username
        local_site = self.local_site

        extra_content = []

        for hook in UserInfoboxHook.hooks:
            try:
                extra_content.append(hook.render(
                    user=user,
                    request=self.request,
                    local_site=local_site))
            except Exception as e:
                logging.exception('Error when running UserInfoboxHook.'
                                  'render method in extension "%s": %s',
                                  hook.extension.id, e)

        review_requests_url = local_site_reverse('user', local_site=local_site,
                                                 args=[username])
        reviews_url = local_site_reverse('user-grid', local_site=local_site,
                                         args=[username, 'reviews'])

        has_avatar = (
            avatar_services.avatars_enabled and
            avatar_services.for_user(user) is not None
        )

        return {
            'extra_content': mark_safe(''.join(extra_content)),
            'full_name': user.get_full_name(),
            'has_avatar': has_avatar,
            'infobox_user': user,
            'review_requests_url': review_requests_url,
            'reviews_url': reviews_url,
            'show_profile': self._show_profile,
            'timezone': self._timezone,
        }


@csrf_protect
def account_register(request, next_url='dashboard'):
    """Display the appropriate registration page.

    If registration is enabled and the selected authentication backend supports
    creation of users, this will return the appropriate registration page. If
    registration is not supported, this will redirect to the login view.
    """
    siteconfig = SiteConfiguration.objects.get_current()
    auth_backends = get_enabled_auth_backends()

    if (auth_backends[0].supports_registration and
            siteconfig.get("auth_enable_registration")):
        response = register(request, next_page=reverse(next_url),
                            form_class=RegistrationForm)

        return response

    return HttpResponseRedirect(reverse("login"))


class MyAccountView(ConfigPagesView):
    """Displays the My Account page containing user preferences.

    The page will be built based on registered pages and forms. This makes
    it easy to plug in new bits of UI for the page, which is handy for
    extensions that want to offer customization for users.
    """

    title = _('My Account')

    css_bundle_names = [
        'account-page',
    ]

    js_bundle_names = [
        '3rdparty-jsonlint',
        'config-forms',
        'account-page',
    ]

    @method_decorator(login_required)
    @augment_method_from(ConfigPagesView)
    def dispatch(self, *args, **kwargs):
        """Handle the view.

        This just falls back to the djblets ConfigPagesView.dispatch
        implementation.
        """
        pass

    @property
    def nav_title(self):
        """Get the title for the navigation section."""
        return self.request.user.username

    @property
    def page_classes(self):
        """The list of page classes for this view.

        If the user is missing any consent requirements or has not accepted
        the privacy policy/terms of service, only the privacy page will be
        shown.
        """
        if self.is_user_missing_consent:
            return [AccountPage.registry.get('page_id', PrivacyPage.page_id)]

        return list(AccountPage.registry)

    @cached_property
    def ordered_user_local_sites(self):
        """Get the user's local sites, ordered by name."""
        return self.request.user.local_site.order_by('name')

    @property
    def render_sidebar(self):
        """Whether or not to render the sidebar.

        If the user is missing any consent requirements or has not accepted
        the privacy policy/terms of service, the sidebar will not render.
        This is to prevent the user from navigating away from the privacy page
        before making decisions.
        """
        return not self.is_user_missing_consent

    @cached_property
    def is_user_missing_consent(self):
        """Whether or not the user is missing consent."""
        return is_consent_missing(self.request.user)


@login_required
@preview_email(prepare_password_changed_mail)
def preview_password_changed_email(request):
    return {
        'user': request.user,
    }


@login_required
@feature_required(oauth2_service_feature)
def edit_oauth_app(request, app_id=None):
    """Create or edit an OAuth2 application.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

        app_id (int, optional):
            The ID of the application to edit.

            If this argument is ``None`` a new application will be edited.

    Returns:
        django.http.HttpResponse:
        The rendered view.
    """
    # If we import this at global scope, it will cause issues with admin sites
    # being automatically registered.
    from reviewboard.oauth.admin import ApplicationAdmin

    if app_id:
        app = get_object_or_404(
            Application,
            pk=app_id,
            user=request.user,
        )
        form_cls = UserApplicationChangeForm
        fieldsets = ApplicationAdmin.fieldsets
    else:
        app = None
        form_cls = UserApplicationCreationForm
        fieldsets = ApplicationAdmin.add_fieldsets

    if request.method == 'POST':
        form_data = request.POST.copy()

        form = form_cls(user=request.user, data=form_data, initial=None,
                        instance=app)

        if form.is_valid():
            app = form.save()

            if app_id is not None:
                next_url = OAuth2Page.get_absolute_url()
            else:
                next_url = reverse('edit-oauth-app', args=(app.pk,))

            return HttpResponseRedirect(next_url)
    else:
        form = form_cls(user=request.user, data=None, initial=None,
                        instance=app)
        # Show a warning at the top of the form when the form is disabled for
        # security.
        #
        # We don't need to worry about full_clean not being called (which would
        # be if we went through form.errors) because this form will never be
        # saved.
        if app and app.is_disabled_for_security:
            form._errors = ErrorDict({
                '__all__': form.error_class(
                    [form.DISABLED_FOR_SECURITY_ERROR],
                ),
            })

    return render_to_response(
        'accounts/edit_oauth_app.html',
        {
            'app': app,
            'form': form,
            'fieldsets': filter_fieldsets(form=form_cls,
                                          fieldsets=fieldsets),
            'oauth2_page_url': OAuth2Page.get_absolute_url(),
            'request': request,
        })



from __future__ import unicode_literals

from django.contrib.auth.models import User
from djblets.webapi.auth import (
    WebAPIBasicAuthBackend as DjbletsWebAPIBasicAuthBackend)
from djblets.webapi.auth.backends.api_tokens import TokenAuthBackendMixin
from djblets.webapi.auth.backends.oauth2_tokens import OAuth2TokenBackendMixin

from reviewboard.accounts.backends import AuthBackend
from reviewboard.webapi.models import WebAPIToken


class TokenAuthBackend(TokenAuthBackendMixin, AuthBackend):
    """Authenticates users and their API tokens for API requests.

    This will handle authenticating users and their API tokens for API
    requests. It's only used for API requests that specify a username and a
    token.
    """

    api_token_model = WebAPIToken


class OAuth2TokenAuthBackend(OAuth2TokenBackendMixin, AuthBackend):
    """An OAuth2 token authentication backend that handles local sites.

    This is similar to :py:class:`oauth2_provider.backends.OAuth2Backend`
    except it ensures the application is enabled and either:

    * not limited to a local site; or
    * limited to the local site being requested.
    """

    def verify_request(self, request, token, user):
        """Ensure the given authentication request is valid.

        This method ensures the following:

        * The Application being used for authentication is enabled.
        * The Local Site the Application is associated with matches the Local
          Site of the current HTTP request.
        * If the Application is associated with a Local Site that site must be
          accessible to the user performing the authentication.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            token (oauth2_provider.models.AccessToken):
                The access token being used for authentication.

            user (django.contrib.auth.models.User):
                The user who is authenticating.

        Returns:
            bool:
            Whether or not the authentication request is valid.
        """
        application = token.application
        return (application.enabled and
                application.local_site == request.local_site and
                (not application.local_site or
                 application.local_site.is_accessible_by(user)))


class WebAPIBasicAuthBackend(DjbletsWebAPIBasicAuthBackend):
    """A specialized WebAPI Basic auth backend that supports e-mail addresses.
    """

    def get_credentials(self, request):
        """Return the credentials supplied in the request.

        If the user provides an e-mail address as the username credential, it
        will be translated to a username.

        Args:
            request (django.http.HttpRequest):
                The request containing the credentials.

        Returns:
            dict:
            A dictionary of the supplied credentials.
        """
        credentials = super(WebAPIBasicAuthBackend, self).get_credentials(
            request)

        if credentials and 'username' in credentials:
            user_exists = (
                User.objects
                .filter(username=credentials['username'])
                .exists()
            )

            if not user_exists and '@' in credentials['username']:
                users = (
                    User.objects
                    .filter(email=credentials['username'])
                    .values_list('username', flat=True)
                )

                if users:
                    credentials['username'] = users[0]

        return credentials



"""Local Site-specfic initialization."""

from __future__ import unicode_literals

from reviewboard.signals import initializing


def _on_initializing(**kwargs):
    """Set up signal handlers for Local Sites."""
    from django.db.models.signals import m2m_changed

    from reviewboard.site.models import LocalSite
    from reviewboard.site.signal_handlers import on_users_changed

    m2m_changed.connect(on_users_changed, sender=LocalSite.users.through)


initializing.connect(_on_initializing)



from __future__ import unicode_literals

from django.conf.urls import include, url

from reviewboard.datagrids import views


urlpatterns = [
    # All Review Requests
    url(r'^r/$',
        views.all_review_requests,
        name='all-review-requests'),

    # Dashboard
    url(r'^dashboard/$',
        views.dashboard,
        name='dashboard'),

    # Users
    url(r'^users/', include([
        url(r'^$',
            views.users_list,
            name='all-users'),
        url(r'^(?P<username>[\w.@+-]+)/$',
            views.submitter,
            name='user'),
        url(r'^(?P<username>[\w.@+-]+)/(?P<grid>[a-z-]+)/$',
            views.submitter,
            name='user-grid'),
    ])),

    # Groups
    url(r'^groups/', include([
        url(r'^$',
            views.group_list,
            name='all-groups'),
        url(r'^(?P<name>[\w-]+)/$',
            views.group,
            name='group'),
        url(r'^(?P<name>[\w-]+)/members/$',
            views.group_members,
            name='group-members'),
    ])),
]



from __future__ import unicode_literals

from datetime import datetime, timedelta
import logging

from django.contrib.auth.models import AnonymousUser, User
from django.utils import timezone
from djblets.testing.decorators import add_fixtures
from djblets.util.dates import get_tz_aware_utcnow
from kgb import SpyAgency, spy_on

from reviewboard.reviews.errors import RevokeShipItError
from reviewboard.reviews.models import Review, ReviewRequest
from reviewboard.reviews.signals import (review_ship_it_revoked,
                                         review_ship_it_revoking)
from reviewboard.testing import TestCase


class ReviewTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.models.Review."""

    fixtures = ['test_users', 'test_scmtools']

    def test_duplicate_reviews(self):
        """Testing consolidation of duplicate reviews"""
        body_top = 'This is the body_top.'
        body_bottom = 'This is the body_bottom.'
        comment_text_1 = 'Comment text 1'
        comment_text_2 = 'Comment text 2'
        comment_text_3 = 'Comment text 3'

        # Some objects we need.
        user = User.objects.get(username='doc')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        # Create the first review.
        master_review = self.create_review(review_request, user=user,
                                           body_top=body_top,
                                           body_bottom='')
        self.create_diff_comment(master_review, filediff, text=comment_text_1,
                                 first_line=1, num_lines=1)

        # Create the second review.
        review = self.create_review(review_request, user=user,
                                    body_top='', body_bottom='')
        self.create_diff_comment(review, filediff, text=comment_text_2,
                                 first_line=1, num_lines=1)

        # Create the third review.
        review = self.create_review(review_request, user=user,
                                    body_top='',
                                    body_bottom=body_bottom)
        self.create_diff_comment(review, filediff, text=comment_text_3,
                                 first_line=1, num_lines=1)

        # Now that we've made a mess, see if we get a single review back.
        logging.disable(logging.WARNING)
        review = review_request.get_pending_review(user)
        self.assertTrue(review)
        self.assertEqual(review.id, master_review.id)
        self.assertEqual(review.body_top, body_top)
        self.assertEqual(review.body_bottom, body_bottom)

        comments = list(review.comments.all())
        self.assertEqual(len(comments), 3)
        self.assertEqual(comments[0].text, comment_text_1)
        self.assertEqual(comments[1].text, comment_text_2)
        self.assertEqual(comments[2].text, comment_text_3)

    def test_is_new_for_user_with_non_owner(self):
        """Testing Review.is_new_for_user with non-owner"""
        user1 = User.objects.create_user(username='test-user-1',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='test-user-2',
                                         email='user2@example.com')

        review = Review(
            user=user1,
            timestamp=datetime(2017, 9, 7, 15, 27, 0))
        self.assertTrue(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 10, 0, 0)))
        self.assertFalse(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 16, 0, 0)))
        self.assertFalse(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 15, 27, 0)))

    def test_is_new_for_user_with_owner(self):
        """Testing Review.is_new_for_user with owner"""
        user = User.objects.create_user(username='test-user',
                                        email='user@example.com')

        review = Review(
            user=user,
            timestamp=datetime(2017, 9, 7, 15, 27, 0))
        self.assertFalse(review.is_new_for_user(
            user=user,
            last_visited=datetime(2017, 9, 7, 16, 0, 0)))

    def test_can_user_revoke_ship_it_with_owner(self):
        """Testing Review.can_user_revoke_ship_it with review owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.assertTrue(review.can_user_revoke_ship_it(review.user))

    def test_can_user_revoke_ship_it_with_non_owner(self):
        """Testing Review.can_user_revoke_ship_it with non-owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.get(username='doc')
        self.assertNotEqual(review.user, user)

        self.assertFalse(review.can_user_revoke_ship_it(user))

    def test_can_user_revoke_ship_it_with_superuser(self):
        """Testing Review.can_user_revoke_ship_it with superuser"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.get(username='admin')
        self.assertNotEqual(review.user, user)

        self.assertTrue(review.can_user_revoke_ship_it(user))

    @add_fixtures(['test_site'])
    def test_can_user_revoke_ship_it_with_local_site_admin(self):
        """Testing Review.can_user_revoke_ship_it with LocalSite admin"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    with_local_site=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.create_user(username='new-site-admin',
                                        email='new_site_admin@example.com')
        review_request.local_site.admins.add(user)
        review_request.local_site.users.add(user)

        self.assertTrue(review.can_user_revoke_ship_it(user))

    def test_can_user_revoke_ship_it_with_anonymous(self):
        """Testing Review.can_user_revoke_ship_it with anonymous user"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.assertFalse(review.can_user_revoke_ship_it(AnonymousUser()))

    def test_can_user_revoke_ship_it_with_unpublished(self):
        """Testing Review.can_user_revoke_ship_it with unpublished review"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)

        self.assertFalse(review.can_user_revoke_ship_it(review.user))

    def test_can_user_revoke_ship_it_with_no_ship_it(self):
        """Testing Review.can_user_revoke_ship_it with no Ship It"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request)

        self.assertFalse(review.can_user_revoke_ship_it(review.user))

    def test_revoke_ship_it(self):
        """Testing Review.revoke_ship_it"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.spy_on(review_ship_it_revoking.send)
        self.spy_on(review_ship_it_revoked.send)

        self.assertEqual(review_request.shipit_count, 1)

        review.revoke_ship_it(review.user)

        # Make sure the signals fired.
        self.assertTrue(review_ship_it_revoking.send.called_with(
            sender=Review, user=review.user, review=review))
        self.assertTrue(review_ship_it_revoked.send.called_with(
            sender=Review, user=review.user, review=review))

        # Check the state of the fields.
        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))
        self.assertEqual(review_request.shipit_count, 0)

        # Make sure they persisted to the database.
        review = Review.objects.get(pk=review.pk)
        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_with_no_ship_it(self):
        """Testing Review.revoke_ship_it with no Ship It"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    publish=True)

        expected_error = 'This review is not marked Ship It!'

        with self.assertRaisesMessage(RevokeShipItError, expected_error):
            review.revoke_ship_it(review.user)

        self.assertEqual(review.body_top, Review.SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)

    def test_revoke_ship_it_with_custom_body_top(self):
        """Testing Review.revoke_ship_it with custom existing body_top"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top='This is a test',
                                    ship_it=True,
                                    publish=True)

        review.revoke_ship_it(review.user)

        self.assertEqual(review.body_top, 'This is a test')
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_with_revoking_signal_exception(self):
        """Testing Review.revoke_ship_it with exception in
        review_ship_it_revoking handler
        """
        def on_revoking(**kwargs):
            raise Exception('oh no')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        try:
            review_ship_it_revoking.connect(on_revoking)

            expected_error = 'Error revoking the Ship It: oh no'

            with self.assertRaisesMessage(RevokeShipItError, expected_error):
                review.revoke_ship_it(review.user)
        finally:
            review_ship_it_revoking.disconnect(on_revoking)

        self.assertEqual(review.body_top, Review.SHIP_IT_TEXT)
        self.assertTrue(review.ship_it)
        self.assertNotIn('revoked_ship_it', review.extra_data)

    def test_revoke_ship_it_with_revoked_signal_exception(self):
        """Testing Review.revoke_ship_it with exception in
        review_ship_it_revoked handler
        """
        def on_revoked(**kwargs):
            raise Exception('oh no')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        try:
            review_ship_it_revoked.connect(on_revoked)
            review.revoke_ship_it(review.user)
        finally:
            review_ship_it_revoked.disconnect(on_revoked)

        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_timestamp(self):
        """Testing Review.revoke_ship_it does not modify the review timestamp
        """
        # ReviewRequest.last_update is a
        # django.db.fields.ModificationTimestampField, which retrieves its
        # value from datetime.utcnow().replace(tzinfo=utc).
        #
        # django.utils.timezone.now has the same implementation.
        #
        # Unfortunately, we cannot spy on datetime.utcnow since it is a
        # builtin. So we replace get_tz_aware_utcnow with timezone.now and we
        # will replace that with a constant function in the spy_on calls below.
        self.spy_on(get_tz_aware_utcnow, call_fake=lambda: timezone.now())

        creation_timestamp = datetime.fromtimestamp(0, timezone.utc)
        review_timestamp = creation_timestamp + timedelta(hours=1)
        revoke_timestamp = review_timestamp + timedelta(hours=1)

        with spy_on(timezone.now, call_fake=lambda: creation_timestamp):
            review_request = self.create_review_request(publish=True)

        with spy_on(timezone.now, call_fake=lambda: review_timestamp):
            review = self.create_review(review_request,
                                        body_top=Review.SHIP_IT_TEXT,
                                        ship_it=True,
                                        publish=True)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)

        self.assertEqual(review_request.time_added, creation_timestamp)
        self.assertEqual(review_request.last_updated, review_timestamp)
        self.assertEqual(review.timestamp, review_timestamp)

        with spy_on(timezone.now, call_fake=lambda: revoke_timestamp):
            review.revoke_ship_it(review.user)

        review = Review.objects.get(pk=review.pk)
        review_request = ReviewRequest.objects.get(pk=review_request.pk)

        self.assertEqual(review_request.time_added, creation_timestamp)
        self.assertEqual(review_request.last_updated, review_timestamp)
        self.assertEqual(review.timestamp, review_timestamp)



"""Signals related to review requests, reviews, and replies."""

from __future__ import unicode_literals

from django.dispatch import Signal


#: Emitted when a review request is publishing.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user publishing the review request.
#:
#:     review_request_draft (reviewboard.reviews.models.ReviewRequestDraft):
#:         The review request draft being published.
review_request_publishing = Signal(providing_args=['user',
                                                   'review_request_draft'])


#: Emitted when a review request is published.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who published the review request.
#:
#:     review_request (reviewboard.reviews.models.ReviewRequest):
#:         The review request that was published.
#:
#:     trivial (bool):
#          Whether or not the review request was published trivially or not.
#:
#:     changedesc (reviewboard.changedescs.models.ChangeDescription):
#:         The change description associated with the publish, if any.
review_request_published = Signal(
    providing_args=['user', 'review_request', 'trivial', 'changedesc'])


#: Emitted when a review request is about to be closed.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user closing the review request.
#:
#:     review_request (reviewboard.reviews.models.ReviewRequest):
#:         The review request being closed.
#:
#:     close_type (unicode):
#:         Describes how the review request is being closed. It is one of
#:         :py:data:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED` or
#:         :py:data:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`.
#:
#:     type (unicode):
#:         Identical to ``close_type``, but deprecated because ``type`` is a
#:         built-in.
#:
#:         .. deprecated:: 3.0
#:            Deprecated in favour of ``close_type``.
#:
#:     description (unicode):
#:         The provided closing description.
#:
#:     rich_text (bool):
#:         Whether or not the description is rich text (Markdown).
review_request_closing = Signal(providing_args=[
    'user', 'review_request',  'close_type', 'type', 'description',
    'rich_text'])


#: Emitted when a review request has been closed.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who closed the review request.
#:
#:     review_request (reviewboard.reviews.models.ReviewRequest):
#:         The review request that was closed.
#:
#:     close_type (unicode):
#:         Describes how the review request was closed. It is one of
#:         :py:data:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED` or
#:         :py:data:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`.
#:
#:     type (unicode):
#:         Identical to ``close_type``, but deprecated because ``type`` is a
#:         builtin.
#:
#:         .. deprecated:: 3.0
#:            Deprecated in favour of ``close_type``.
#:
#:     description (unicode):
#:         The provided closing description.
#:
#:     rich_text (bool):
#:         Whether or not the description is rich text (Markdown).
review_request_closed = Signal(providing_args=['user', 'review_request',
                                               'type', 'description',
                                               'rich_text'])


#: Emitted when a review request is about to be reopened.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user re-opening the review request.
#:
#:     review_request (reviewboard.reviews.models.ReviewRequest):
#:         The review request being reopened.
review_request_reopening = Signal(providing_args=['user', 'review_request'])


#: Emitted when a review request has been reopened.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who re-opened the review request.
#:
#:     review_request (reviewboard.reviews.models.ReviewRequest):
#:         The review request that was reopened.
#:
#:     old_status (unicode):
#:         The old status for the review request. This will be
#:         :py:attr:`~reviewboard.reviews.models.ReviewRequest.PENDING_REVIEW`,
#:         :py:attr:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED`, or
#:         :py:attr:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`.
#:
#:     old_public (bool):
#:         The old public state for the review request.
review_request_reopened = Signal(providing_args=['user', 'review_request',
                                                 'old_status', 'old_public'])


#: Emitted when a review is being published.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user publishing the review.
#:
#:     review (reviewboard.reviews.models.Review):
#:         The review that's being published.
#:
#:     to_owner_only (boolean):
#:         Whether the review e-mail should be sent only to the review request
#:         submitter.
review_publishing = Signal(providing_args=['user', 'review',
                                           'to_owner_only'])


#: Emitted when a Ship It is about to be revoked from a review.
#:
#: Listeners can raise a
#: :py:exc:`~reviewboard.reviews.errors.RevokeShipItError` to stop the Ship It
#: from being revoking.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who requested to revoke the Ship It.
#:
#:     review (reviewboard.reviews.models.review.Review):
#:         The review that will have its Ship It revoked.
review_ship_it_revoking = Signal(providing_args=['user', 'review'])


#: Emitted when a Ship It has been revoked from a review.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who revoked the Ship It.
#:
#:     review (reviewboard.reviews.models.review.Review):
#:         The review that had its Ship It revoked.
review_ship_it_revoked = Signal(providing_args=['user', 'review'])


#: Emitted when a review has been published.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who published the review.
#:
#:     review (reviewboard.reviews.models.Review):
#:         The review that was published.
#:
#:     to_owner_only (boolean):
#:         Whether the review e-mail should be sent only to the review request
#:         submitter.
#:
#:     request (django.http.HttpRequest):
#:         The request object if the review was published from an HTTP request.
review_published = Signal(
    providing_args=['user', 'review', 'to_owner_only', 'request'])


#: Emitted when a reply to a review is being published.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user publishing the reply.
#:
#:     review (reviewboard.reviews.models.Review):
#:         The reply that's being published.
reply_publishing = Signal(providing_args=['user', 'reply'])


#: Emitted when a reply to a review has ben published.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user who published the reply.
#:
#:     review (reviewboard.reviews.models.Review):
#:         The reply that was published.
#:
#:     trivial (bool):
#:         Whether the reply was considered trivial.
reply_published = Signal(providing_args=['user', 'reply', 'trivial'])



"""Privacy support for user accounts."""

from __future__ import unicode_literals

from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _

from djblets.privacy.consent import (Consent,
                                     get_consent_requirements_registry,
                                     get_consent_tracker)
from djblets.privacy.consent.common import (BaseGravatarConsentRequirement,
                                            PolicyConsentRequirement)
from djblets.privacy.consent.registry import ConsentRequirementsRegistry
from djblets.registries.errors import ItemLookupError
from djblets.registries.signals import registry_populating
from djblets.siteconfig.models import SiteConfiguration
from djblets.urls.staticfiles import static_lazy
from djblets.util.html import mark_safe_lazy


class GravatarConsentRequirement(BaseGravatarConsentRequirement):
    intent_description = mark_safe_lazy(_(
        "Gravatar is used by many services and applications to manage and "
        "serve <b>avatars</b>. We use Gravatars by default for your "
        "avatar. If you don't want to use it, you can block that here, "
        "or upload your own avatar in your Profile settings."
    ))

    icons = {
        '1x': static_lazy('rb/images/consent/gravatar.png'),
        '2x': static_lazy('rb/images/consent/gravatar@2x.png'),
    }


def recompute_privacy_consents(unregister=True):
    """Recompute the built-in consent requirements for user privacy.

    Args:
        unregister (bool, optional):
            Whether to unregister previous entries first.
    """
    registry = get_consent_requirements_registry()

    if unregister:
        # Unregister our consent requirements (but leave ones provided by
        # extensions).
        for requirement in (GravatarConsentRequirement,
                            PolicyConsentRequirement):
            try:
                registry.unregister_by_attr(
                    'requirement_id', requirement.requirement_id)
            except ItemLookupError:
                pass

    siteconfig = SiteConfiguration.objects.get_current()
    privacy_policy = siteconfig.get('privacy_policy_url')
    terms_of_service = siteconfig.get('terms_of_service_url')

    if privacy_policy or terms_of_service:
        registry.register(PolicyConsentRequirement(
            privacy_policy,
            terms_of_service,
            siteconfig.get('site_admin_email')))

    registry.register(GravatarConsentRequirement())


def is_consent_missing(user):
    """Return whether the user is missing any consent requirements.

    Args:
        user (django.contrib.auth.models.User):
            The user in question.

    Returns:
        bool:
        Whether or not the user is missing any consent requirements.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if not siteconfig.get('privacy_enable_user_consent'):
        return False

    consent_tracker = get_consent_tracker()
    pending_consent = consent_tracker.get_pending_consent_requirements(
        user)

    needs_accept_policies = (
        (siteconfig.get('privacy_policy_url') or
         siteconfig.get('terms_of_service_url')) and
        (consent_tracker.get_consent(
            user,
            PolicyConsentRequirement.requirement_id) !=
         Consent.GRANTED)
    )

    return needs_accept_policies or pending_consent


@receiver(registry_populating, sender=ConsentRequirementsRegistry)
def _on_consent_requirements_registry_populating(**kwargs):
    """Handler for when the requirements registry is populating.

    This will force a recomputation of the privacy consents, registering
    new ones based on the current settings.

    Args:
        **kwargs (dict):
            Keyword arguments passed to the handler.
    """
    recompute_privacy_consents(unregister=False)



from __future__ import unicode_literals

import re

from django.contrib.auth.models import User
from djblets.conditions import ConditionSet, Condition

from reviewboard.reviews.conditions import (AnyReviewGroupsPublicOperator,
                                            AllReviewGroupsInviteOnlyOperator,
                                            ReviewGroupsChoice,
                                            ReviewRequestAllDiffFilesChoice,
                                            ReviewRequestAnyDiffFileChoice,
                                            ReviewRequestRepositoriesChoice,
                                            ReviewRequestRepositoryTypeChoice,
                                            ReviewRequestReviewGroupsChoice,
                                            ReviewRequestOwnerChoice,
                                            ReviewRequestReviewerChoice,
                                            ReviewRequestParticipantChoice)
from reviewboard.reviews.models import Group
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class ReviewGroupOperatorTests(TestCase):
    """Unit tests for review group condition operators."""

    def test_any_public_with_match(self):
        """Testing AnyReviewGroupsPublicOperator with match"""

        self.assertTrue(self._check_match(
            AnyReviewGroupsPublicOperator,
            [
                self.create_review_group(name='group1', invite_only=False),
                self.create_review_group(name='group2', invite_only=True),
            ]))

    def test_any_public_without_match(self):
        """Testing AnyReviewGroupsPublicOperator without match"""
        self.create_review_group(name='group1', invite_only=False)
        private_group = self.create_review_group(name='group2',
                                                 invite_only=True)

        self.assertFalse(self._check_match(
            AnyReviewGroupsPublicOperator,
            [private_group]))
        self.assertFalse(self._check_match(
            AnyReviewGroupsPublicOperator,
            []))

    def test_all_invite_only_with_match(self):
        """Testing AllReviewGroupsInviteOnlyOperator with match"""

        self.assertTrue(self._check_match(
            AllReviewGroupsInviteOnlyOperator,
            [
                self.create_review_group(name='group1', invite_only=True),
                self.create_review_group(name='group2', invite_only=True),
            ]))

    def test_all_invite_only_without_match(self):
        """Testing AllReviewGroupsInviteOnlyOperator without match"""
        group1 = self.create_review_group(name='group1', invite_only=False)
        group2 = self.create_review_group(name='group2', invite_only=True)

        self.assertFalse(self._check_match(
            AllReviewGroupsInviteOnlyOperator,
            [group1, group2]))
        self.assertFalse(self._check_match(
            AllReviewGroupsInviteOnlyOperator,
            []))

    def _check_match(self, op_cls, match_value, condition_value=None):
        op = op_cls(None)
        return op.matches(match_value=match_value,
                          condition_value=condition_value)


class ReviewGroupsChoiceTests(TestCase):
    """Unit tests for reviewboard.reviews.conditions.ReviewGroupsChoice."""

    def setUp(self):
        super(ReviewGroupsChoiceTests, self).setUp()

        self.choice = ReviewGroupsChoice()

    def test_get_queryset(self):
        """Testing ReviewGroupsChoice.get_queryset"""
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [group1.pk, group2.pk],
            transform=lambda group: group.pk)

    def test_get_queryset_with_local_site(self):
        """Testing ReviewGroupsChoice.get_queryset with LocalSite"""
        good_site = LocalSite.objects.create(name='good-site')
        bad_site = LocalSite.objects.create(name='bad-site')

        # These should match.
        group1 = self.create_review_group(name='group1',
                                          local_site=good_site)
        group2 = self.create_review_group(name='group2',
                                          local_site=good_site)

        # These should not match.
        self.create_review_group(name='group3')
        self.create_review_group(name='group4', local_site=bad_site)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [group1.pk, group2.pk],
            transform=lambda group: group.pk)

    def test_matches_with_any_op(self):
        """Testing ReviewGroupsChoice.matches with "any" operator"""
        self.create_review_group(name='group1', invite_only=False)
        self.create_review_group(name='group2', invite_only=True)

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any')),
        ])

        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.all()))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.none()))

    def test_matches_with_none_op(self):
        """Testing ReviewGroupsChoice.matches with "none" operator"""
        self.create_review_group(name='group1')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('none')),
        ])

        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.none()))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.all()))

    def test_matches_with_contains_any_op(self):
        """Testing ReviewGroupsChoice.matches with "contains-any" operator"""
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        group3 = self.create_review_group(name='group3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains-any'),
                      [group1, group2])
        ])

        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.filter(pk=group1.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.filter(pk=group3.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.none()))

    def test_matches_with_does_not_contain_any_op(self):
        """Testing ReviewGroupsChoice.matches with "does-not-contain-any"
        operator
        """
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        group3 = self.create_review_group(name='group3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain-any'),
                      [group1, group2])
        ])

        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.filter(pk=group1.pk)))
        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.filter(pk=group3.pk)))
        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.none()))

    def test_matches_with_any_public_op(self):
        """Testing ReviewGroupsChoice.matches with "any-public" operator"""
        group1 = self.create_review_group(name='group1', invite_only=False)
        group2 = self.create_review_group(name='group2', invite_only=True)

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any-public')),
        ])

        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.filter(pk=group1.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.filter(pk=group2.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.none()))

    def test_matches_with_all_invite_only_op(self):
        """Testing ReviewGroupsChoice.matches with "all-invite-only" operator
        """
        group1 = self.create_review_group(name='group1', invite_only=True)
        group2 = self.create_review_group(name='group2', invite_only=False)

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('all-invite-only')),
        ])

        self.assertTrue(condition_set.matches(
            review_groups=Group.objects.filter(pk=group1.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.filter(pk=group2.pk)))
        self.assertFalse(condition_set.matches(
            review_groups=Group.objects.none()))


class ReviewRequestAllDiffFilesChoiceTests(TestCase):
    """Unit tests for ReviewRequestAllDiffFilesChoice."""

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(ReviewRequestAllDiffFilesChoiceTests, self).setUp()

        self.choice = ReviewRequestAllDiffFilesChoice()
        self.review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(self.review_request)
        self.filediff1 = self.create_filediff(diffset, source_file='file1',
                                              dest_file='file1')
        self.filediff2 = self.create_filediff(diffset, source_file='file2',
                                              dest_file='file2')

    def test_get_match_value(self):
        """Testing ReviewRequestAllDiffFilesChoice.get_match_value"""
        self.assertEqual(self.choice.get_match_value(self.review_request, {}),
                         {'file1', 'file2'})

    def test_matches_with_is_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "is" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is'), 'file1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

        self.filediff2.delete()
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_is_not_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "is-not"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-not'),
                      'fileX'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-not'),
                      'file1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_contains_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "contains"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains'),
                      'file'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains'),
                      '1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_does_not_contain_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with
        "does-not-contain" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain'),
                      '3'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain'),
                      'ile1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_starts_with_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "starts-with"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('starts-with'),
                      'file'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('starts-with'),
                      'file1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_ends_with_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "ends-with"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('ends-with'),
                      'le1'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

        self.filediff2.delete()
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_matches_regex_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with "matches-regex"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('matches-regex'),
                      re.compile(r'^[Ff]ile\d$')),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('matches-regex'),
                      re.compile('^[Ff]ile1$')),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_does_not_match_regex_op(self):
        """Testing ReviewRequestAllDiffFilesChoice.matches with
        "does-not-match-regex" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-match-regex'),
                      re.compile(r'^[Ff]ile3$')),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-match-regex'),
                      re.compile('[Ff]ile1')),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))


class ReviewRequestAnyDiffFileChoiceTests(TestCase):
    """Unit tests for ReviewRequestAnyDiffFileChoice."""

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(ReviewRequestAnyDiffFileChoiceTests, self).setUp()

        self.choice = ReviewRequestAnyDiffFileChoice()
        self.review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(self.review_request)
        self.filediff1 = self.create_filediff(diffset, source_file='file1',
                                              dest_file='file1')
        self.filediff2 = self.create_filediff(diffset, source_file='file2',
                                              dest_file='file2')

    def test_get_match_value(self):
        """Testing ReviewRequestAnyDiffFileChoice.get_match_value"""
        self.assertEqual(self.choice.get_match_value(self.review_request, {}),
                         {'file1', 'file2'})

    def test_matches_with_is_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "is" operator"""
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is'), 'file2'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is'), 'fileX'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_is_not_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "is-not"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-not'),
                      'fileX'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-not'),
                      'file1'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        self.filediff2.delete()
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_contains_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "contains"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains'),
                      '1'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains'),
                      '3'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_does_not_contain_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with
        "does-not-contain" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain'),
                      'xyz'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain'),
                      'file'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_starts_with_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "starts-with"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('starts-with'),
                      'file'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('starts-with'),
                      'ile'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_ends_with_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "ends-with"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('ends-with'),
                      'le1'),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('ends-with'),
                      'xyz'),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_matches_regex_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with "matches-regex"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('matches-regex'),
                      re.compile(r'^[Ff]ile1$')),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('matches-regex'),
                      re.compile('^\d')),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))

    def test_matches_with_does_not_match_regex_op(self):
        """Testing ReviewRequestAnyDiffFileChoice.matches with
        "does-not-match-regex" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-match-regex'),
                      re.compile(r'^\d')),
        ])
        self.assertTrue(condition_set.matches(
            review_request=self.review_request))

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-match-regex'),
                      re.compile('[Ff]ile\d')),
        ])
        self.assertFalse(condition_set.matches(
            review_request=self.review_request))


class ReviewRequestRepositoriesChoiceTests(TestCase):
    """Unit tests for ReviewRequestRepositoriesChoice."""

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(ReviewRequestRepositoriesChoiceTests, self).setUp()

        self.choice = ReviewRequestRepositoriesChoice()

    def test_matches_with_any_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "any" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any')),
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(create_repository=True)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request()))

    def test_matches_with_none_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "none" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('none')),
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request()))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(create_repository=True)))

    def test_matches_with_one_of_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "one-of"
        operator
        """
        repository1 = self.create_repository(name='repo1')
        repository2 = self.create_repository(name='repo2')
        repository3 = self.create_repository(name='repo3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('one-of'),
                      [repository1, repository2])
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository1)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository2)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository3)))

    def test_matches_with_not_one_of_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "not-one-of"
        operator
        """
        repository1 = self.create_repository(name='repo1')
        repository2 = self.create_repository(name='repo2')
        repository3 = self.create_repository(name='repo3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('not-one-of'),
                      [repository1, repository2])
        ])

        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository1)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository2)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository3)))

    def test_matches_with_is_public_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "is-public"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-public')),
        ])

        public_repo = self.create_repository(name='repo1', public=True)
        private_repo = self.create_repository(name='repo2', public=False)

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=public_repo)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                repository=private_repo)))

    def test_matches_with_is_private_op(self):
        """Testing ReviewRequestRepositoriesChoice.matches with "is-private"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-private')),
        ])

        public_repo = self.create_repository(name='repo1', public=True)
        private_repo = self.create_repository(name='repo2', public=False)

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                repository=private_repo)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=public_repo)))


class ReviewRequestRepositoryTypeChoiceTests(TestCase):
    """Unit tests for ReviewRequestRepositoryTypeChoice."""

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        super(ReviewRequestRepositoryTypeChoiceTests, self).setUp()

        self.choice = ReviewRequestRepositoryTypeChoice()

    def test_matches_with_one_of_op(self):
        """Testing ReviewRequestRepositoryTypeChoice.matches with "one-of"
        operator
        """
        repository1 = self.create_repository(name='repo1',
                                             tool_name='Git')
        repository2 = self.create_repository(name='repo2',
                                             tool_name='Subversion')
        repository3 = self.create_repository(name='repo3',
                                             tool_name='CVS')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('one-of'),
                      [repository1.tool, repository2.tool])
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository1)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository2)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository3)))

    def test_matches_with_not_one_of_op(self):
        """Testing ReviewRequestRepositoryTypeChoice.matches with "not-one-of"
        operator
        """
        repository1 = self.create_repository(name='repo1',
                                             tool_name='Git')
        repository2 = self.create_repository(name='repo2',
                                             tool_name='Subversion')
        repository3 = self.create_repository(name='repo3',
                                             tool_name='CVS')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('not-one-of'),
                      [repository1.tool, repository2.tool])
        ])

        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository1)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(repository=repository2)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(repository=repository3)))


class ReviewRequestReviewGroupsChoiceTests(TestCase):
    """Unit tests for ReviewRequestReviewGroupsChoice."""

    fixtures = ['test_users']

    def setUp(self):
        super(ReviewRequestReviewGroupsChoiceTests, self).setUp()

        self.choice = ReviewRequestReviewGroupsChoice()

    def test_get_queryset(self):
        """Testing ReviewGroupsChoice.get_queryset"""
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [group1.pk, group2.pk],
            transform=lambda group: group.pk)

    def test_get_queryset_with_local_site(self):
        """Testing ReviewGroupsChoice.get_queryset with LocalSite"""
        good_site = LocalSite.objects.create(name='good-site')
        bad_site = LocalSite.objects.create(name='bad-site')

        # These should match.
        group1 = self.create_review_group(name='group1',
                                          local_site=good_site)
        group2 = self.create_review_group(name='group2',
                                          local_site=good_site)

        # These should not match.
        self.create_review_group(name='group3')
        self.create_review_group(name='group4', local_site=bad_site)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [group1.pk, group2.pk],
            transform=lambda group: group.pk)

    def test_matches_with_any_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with "any" operator
        """
        group = self.create_review_group(name='group1')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any')),
        ])

        review_request = self.create_review_request()
        review_request.target_groups = [group]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = []
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_none_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with "none" operator
        """
        group = self.create_review_group(name='group1')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('none')),
        ])

        review_request = self.create_review_request()
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group]
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_contains_any_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with "contains-any"
        operator
        """
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        group3 = self.create_review_group(name='group2')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('contains-any'),
                      [group1, group2]),
        ])

        review_request = self.create_review_request()
        review_request.target_groups = [group2]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group3]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request.target_groups = []
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_does_not_contain_any_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with
        "does-not-contain-any" operator
        """
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        group3 = self.create_review_group(name='group3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain-any'),
                      [group1, group2])
        ])

        review_request = self.create_review_request()
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group3]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group1]
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_any_public_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with "any-public"
        operator"""
        group1 = self.create_review_group(name='group1', invite_only=False)
        group2 = self.create_review_group(name='group2', invite_only=True)

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any-public')),
        ])

        review_request = self.create_review_request()
        review_request.target_groups = [group1]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group2]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request.target_groups = []
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_all_invite_only_op(self):
        """Testing ReviewRequestReviewGroupsChoice.matches with
        "all-invite-only" operator
        """
        group1 = self.create_review_group(name='group1', invite_only=True)
        group2 = self.create_review_group(name='group2', invite_only=False)

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('all-invite-only')),
        ])

        review_request = self.create_review_request()
        review_request.target_groups = [group1]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request.target_groups = [group2]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request.target_groups = []
        self.assertFalse(condition_set.matches(review_request=review_request))


class ReviewRequestOwnerChoiceTests(TestCase):
    """Unit tests for ReviewRequestOwnerChoice."""

    fixtures = ['test_users']

    def setUp(self):
        super(ReviewRequestOwnerChoiceTests, self).setUp()

        self.choice = ReviewRequestOwnerChoice()
        self.user1 = User.objects.get(username='doc')
        self.user2 = User.objects.get(username='grumpy')
        self.user3 = User.objects.get(username='dopey')

    def test_get_queryset(self):
        """Testing ReviewRequestOwnerChoice.get_queryset"""
        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            User.objects.values_list('pk', flat=True),
            transform=lambda user: user.pk)

    def test_get_queryset_with_local_site(self):
        """Testing ReviewRequestOwnerChoice.get_queryset with LocalSite"""
        good_site = LocalSite.objects.create(name='good-site')
        good_site.users.add(self.user2)

        bad_site = LocalSite.objects.create(name='bad-site')
        bad_site.users.add(self.user3)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [self.user2.pk],
            transform=lambda user: user.pk)

    def test_matches_with_one_of_op(self):
        """Testing ReviewRequestOwnerChoice.matches with "one-of"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('one-of'),
                      [self.user1, self.user2]),
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user1)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user2)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user3)))

    def test_matches_with_not_one_of_op(self):
        """Testing ReviewRequestOwnerChoice.matches with "not-one-of"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('not-one-of'),
                      [self.user1, self.user2]),
        ])

        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user1)))
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user2)))
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(submitter=self.user3)))


class ReviewRequestReviewerChoiceTests(TestCase):
    """Unit tests for ReviewRequestReviewerChoice."""

    fixtures = ['test_users']

    def setUp(self):
        super(ReviewRequestReviewerChoiceTests, self).setUp()

        self.choice = ReviewRequestReviewerChoice()
        self.user1 = User.objects.get(username='doc')
        self.user2 = User.objects.get(username='grumpy')
        self.user3 = User.objects.get(username='dopey')

    def test_get_queryset(self):
        """Testing ReviewRequestReviewerChoice.get_queryset"""
        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            User.objects.values_list('pk', flat=True),
            transform=lambda user: user.pk)

    def test_get_queryset_with_local_site(self):
        """Testing ReviewRequestReviewerChoice.get_queryset with LocalSite"""
        good_site = LocalSite.objects.create(name='good-site')
        good_site.users.add(self.user2)

        bad_site = LocalSite.objects.create(name='bad-site')
        bad_site.users.add(self.user3)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [self.user2.pk],
            transform=lambda user: user.pk)

    def test_matches_with_contains_any_op(self):
        """Testing ReviewRequestReviewerChoice.matches with
        "contains-any" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('contains-any'),
                      [self.user1, self.user2]),
        ])

        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user2]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1, self.user2]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user2, self.user1]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1, self.user3]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3, self.user1]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1, self.user2, self.user3]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3, self.user2, self.user1]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3, self.user1, self.user2]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3]))
        )

    def test_matches_with_does_not_contain_any_op(self):
        """Testing ReviewRequestReviewerChoice.matches with
        "does-not-contain-any" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain-any'),
                      [self.user1, self.user2]),
        ])

        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user2]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1, self.user2]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user2, self.user1]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user1, self.user2, self.user3]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3, self.user2, self.user1]))
        )
        self.assertFalse(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3, self.user1, self.user2]))
        )
        self.assertTrue(condition_set.matches(
            review_request=self.create_review_request(
                target_people=[self.user3]))
        )


class ReviewRequestParticipantChoiceTests(TestCase):
    """Unit tests for ReviewRequestParticipantChoice."""

    fixtures = ['test_users']

    def setUp(self):
        super(ReviewRequestParticipantChoiceTests, self).setUp()

        self.choice = ReviewRequestParticipantChoice()
        self.user1 = User.objects.get(username='doc')
        self.user2 = User.objects.get(username='grumpy')
        self.user3 = User.objects.get(username='dopey')

    def test_get_queryset(self):
        """Testing ReviewRequestParticipantChoice.get_queryset"""
        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            User.objects.values_list('pk', flat=True),
            transform=lambda user: user.pk)

    def test_get_queryset_with_local_site(self):
        """Testing ReviewRequestParticipantChoice.get_queryset with
        LocalSite
        """
        good_site = LocalSite.objects.create(name='good-site')
        good_site.users.add(self.user2)

        bad_site = LocalSite.objects.create(name='bad-site')
        bad_site.users.add(self.user3)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset(),
            [self.user2.pk],
            transform=lambda user: user.pk)

    def test_matches_with_contains_any_op(self):
        """Testing ReviewRequestParticipantChoice.matches with "contains-any"
        operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('contains-any'),
                      [self.user1, self.user2]),
        ])

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user3)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user3)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user3)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

    def test_matches_with_does_not_contain_any_op(self):
        """Testing ReviewRequestParticipantChoice.matches with
        "does-not-contain-any" operator
        """
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice,
                      self.choice.get_operator('does-not-contain-any'),
                      [self.user1, self.user2]),
        ])

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user3)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user3)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user1),
            self.create_review(review_request, user=self.user2)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3),
            self.create_review(review_request, user=self.user2),
            self.create_review(review_request, user=self.user1)
        ]
        self.assertFalse(condition_set.matches(review_request=review_request))

        review_request = self.create_review_request()
        review_request.reviews = [
            self.create_review(review_request, user=self.user3)
        ]
        self.assertTrue(condition_set.matches(review_request=review_request))



from __future__ import unicode_literals

import warnings

import pymdownx.emoji
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from django.utils.html import escape
from djblets import markdown as djblets_markdown
from djblets.siteconfig.models import SiteConfiguration
from markdown import markdown

from reviewboard.deprecation import RemovedInReviewBoard40Warning


# Keyword arguments used when calling a Markdown renderer function.
#
# We use XHTML1 instead of HTML5 to ensure the results can be parsed by an
# XML parser, needed for change descriptions and other parts of the web UI.
MARKDOWN_KWARGS = {
    'enable_attributes': False,
    'output_format': 'xhtml1',
    'lazy_ol': False,
    'extensions': [
        'markdown.extensions.fenced_code',
        'markdown.extensions.codehilite',
        'markdown.extensions.sane_lists',
        'markdown.extensions.smart_strong',
        'markdown.extensions.tables',
        'markdown.extensions.nl2br',
        'pymdownx.tilde',
        'pymdownx.emoji',
        'djblets.markdown.extensions.escape_html',
        'djblets.markdown.extensions.wysiwyg',
    ],
    'extension_configs': {
        'codehilite': {
            'guess_lang': False,
        },
        'pymdownx.emoji': {
            'emoji_index': pymdownx.emoji.gemoji,
            'options': {
                'classes': 'emoji',
            },
        },
    },
}


def markdown_escape(text):
    """Escapes text for use in Markdown.

    This will escape the provided text so that none of the characters will
    be rendered specially by Markdown.

    This is deprecated. Please use djblets.markdown.markdown_escape instead.
    """
    warnings.warn('reviewboard.reviews.markdown_utils.markdown_escape is '
                  'deprecated. Please use djblets.markdown.markdown_escape.',
                  RemovedInReviewBoard40Warning)

    return djblets_markdown.markdown_escape(text)


def markdown_unescape(escaped_text):
    """Unescapes Markdown-escaped text.

    This will unescape the provided Markdown-formatted text so that any
    escaped characters will be unescaped.

    This is deprecated. Please use djblets.markdown.markdown_unescape instead.
    """
    warnings.warn('reviewboard.reviews.markdown_utils.markdown_unescape is '
                  'deprecated. Please use djblets.markdown.markdown_unescape.',
                  RemovedInReviewBoard40Warning)

    return djblets_markdown.markdown_unescape(escaped_text)


def markdown_escape_field(obj, field_name):
    """Escapes Markdown text in a model or dictionary's field.

    This is a convenience around markdown_escape to escape the contents of
    a particular field in a model or dictionary.
    """
    if isinstance(obj, Model):
        setattr(obj, field_name,
                djblets_markdown.markdown_escape(getattr(obj, field_name)))
    elif isinstance(obj, dict):
        obj[field_name] = djblets_markdown.markdown_escape(obj[field_name])
    else:
        raise TypeError('Unexpected type %r passed to markdown_escape_field'
                        % obj)


def markdown_unescape_field(obj, field_name):
    """Unescapes Markdown text in a model or dictionary's field.

    This is a convenience around markdown_unescape to unescape the contents of
    a particular field in a model or dictionary.
    """
    if isinstance(obj, Model):
        setattr(obj, field_name, markdown_unescape(getattr(obj, field_name)))
    elif isinstance(obj, dict):
        obj[field_name] = markdown_unescape(obj[field_name])
    else:
        raise TypeError('Unexpected type %r passed to markdown_unescape_field'
                        % obj)


def normalize_text_for_edit(user, text, rich_text, escape_html=True):
    """Normalizes text, converting it for editing.

    This will normalize text for editing based on the rich_text flag and
    the user settings.

    If the text is not in Markdown and the user edits in Markdown by default,
    this will return the text escaped for edit. Otherwise, the text is
    returned as-is.
    """
    if text is None:
        return ''

    if not rich_text and is_rich_text_default_for_user(user):
        # This isn't rich text, but it's going to be edited as rich text,
        # so escape it.
        text = djblets_markdown.markdown_escape(text)

    if escape_html:
        text = escape(text)

    return text


def markdown_render_conditional(text, rich_text):
    """Return the escaped HTML content based on the rich_text flag."""
    if rich_text:
        return render_markdown(text)
    else:
        return escape(text)


def is_rich_text_default_for_user(user):
    """Returns whether the user edits in Markdown by default."""
    if user.is_authenticated():
        try:
            return user.get_profile().should_use_rich_text
        except ObjectDoesNotExist:
            pass

    siteconfig = SiteConfiguration.objects.get_current()

    return siteconfig.get('default_use_rich_text')


def markdown_set_field_escaped(obj, field, escaped):
    """Escapes or unescapes the specified field in a model or dictionary."""
    if escaped:
        markdown_escape_field(obj, field)
    else:
        markdown_unescape_field(obj, field)


def iter_markdown_lines(markdown_html):
    """Iterates over lines of Markdown, normalizing for individual display.

    Generated Markdown HTML cannot by itself be handled on a per-line-basis.
    Code blocks, for example, will consist of multiple lines of content
    contained within a <pre> tag. Likewise, lists will be a bunch of
    <li> tags inside a <ul> tag, and individually do not form valid lists.

    This function iterates through the Markdown tree and generates
    self-contained lines of HTML that can be rendered individually.

    This is deprecated. Please use djblets.markdown.iter_markdown_lines
    instead.
    """
    warnings.warn(
        'reviewboard.reviews.markdown_utils.iter_markdown_lines is '
        'deprecated. Please use djblets.markdown.iter_markdown_lines.',
        RemovedInReviewBoard40Warning)

    return djblets_markdown.iter_markdown_lines(markdown_html)


def get_markdown_element_tree(markdown_html):
    """Returns an XML element tree for Markdown-generated HTML.

    This will build the tree and return all nodes representing the rendered
    Markdown content.

    This is deprecated. Please use djblets.markdown.get_markdown_element_tree
    instead.
    """
    warnings.warn(
        'reviewboard.reviews.markdown_utils.get_markdown_element_tree is '
        'deprecated. Please use djblets.markdown.get_markdown_element_tree.',
        RemovedInReviewBoard40Warning)

    return djblets_markdown.get_markdown_element_tree(markdown_html)


def sanitize_illegal_chars_for_xml(s):
    """Sanitize a string, removing characters illegal in XML.

    This will remove a number of characters that would break the  XML parser.
    They may be in the string due to a copy/paste.

    This code is courtesy of the XmlRpcPlugin developers, as documented
    here: http://stackoverflow.com/a/22273639

    This is deprecated. Please use
    djblets.markdown.sanitize_illegal_chars_for_xml instead.
    """
    warnings.warn(
        'reviewboard.reviews.markdown_utils.sanitize_illegal_chars_for_xml '
        'is deprecated. Please use '
        'djblets.markdown.sanitize_illegal_chars_for_xml.',
        RemovedInReviewBoard40Warning)

    return djblets_markdown.sanitize_illegal_chars_for_xml(s)


def render_markdown(text):
    """Renders Markdown text to HTML.

    The Markdown text will be sanitized to prevent injecting custom HTML.
    It will also enable a few plugins for code highlighting and sane lists.
    """
    if isinstance(text, bytes):
        text = text.decode('utf-8')

    return markdown(text, **MARKDOWN_KWARGS)


def render_markdown_from_file(f):
    """Renders Markdown text to HTML.

    The Markdown text will be sanitized to prevent injecting custom HTML.
    It will also enable a few plugins for code highlighting and sane lists.
    """
    return djblets_markdown.render_markdown_from_file(f, **MARKDOWN_KWARGS)
